// *F* Developed 2011-2014 rejecht <rejecht at outlook dot com>

class Award extends Object
	abstract
	config (User)
	;

// #

var const Color AwardMultiplierColor;

var(Colors) GlobalConfig Color
	CaptionColor,
	CyanCaptionColor,
	GrayCaptionColor,
	OrangeCaptionColor,
	PurpleCaptionColor,
	RedCaptionColor,
	WhiteCaptionColor;

// #Queries

static function string GetCaptionString (int AwardIndex)
{
	return "";
}

static function Color GetCaptionColor (int AwardIndex)
{
	return Default.CaptionColor;
}

// #

// - Check if PlayerID matches that of the currently spectated player.

static final function bool IsSpectatingPlayerID (PlayerController PC, int PlayerID)
{
	local Pawn SpectatedPawn;

	SpectatedPawn = Pawn(PC.ViewTarget);

	if (SpectatedPawn == None)
	{
		SpectatedPawn = PC.Pawn;
	}

	if
	(
		(SpectatedPawn != None)
		&&
		(SpectatedPawn.PlayerReplicationInfo != None)
	)
	{
		return (SpectatedPawn.PlayerReplicationInfo.PlayerID == PlayerID);
	}
	else
	{
		return False;
	}
}

// #

// Allows overriding to add more flashy messages.
// 1. Some player gets awarded.
// 2. Server sends a message to all players who have the player as a ViewTarget, which always includes the player itself.
// 3. Due to latency, client always checks viewed PlayerID for match with PlayerID in message and calls this function on a match.
// 4. Players who view the awarded player all see the same award.
// - If a message is not overridden, it will show at the top part of the screen.

protected static function HandleLocalMessage
(
	PlayerController PC,
	PlayerReplicationInfo RelatedPRI_1,
	int AwardIndex
)
{
	PC.ReceiveLocalizedMessage (Class'LocalMessage_Award', AwardIndex, RelatedPRI_1, None, Default.Class);
}

// #Dispatch

static function ClientReceive
(
	PlayerController PC,
	optional int AwardIndex,
	optional int Points, // (Points > 0)
	optional int Multiplier, // (Multiplier > 1)
	optional int PlayerID
)
{
	if
	(
		Class'PlayerSettings'.Default.bShowAwardStatusAnnouncements
		&&
		(PC.GameReplicationInfo != None)
		&&
		IsSpectatingPlayerID (PC, PlayerID) // Is client spectating?
	)
	{
		HandleLocalMessage (PC, PC.GameReplicationInfo.FindPlayerByID (PlayerID), AwardIndex);
	}

	// Console message for awards given to the local player.

	if
	(
		(PC.PlayerReplicationInfo != None)
		&&
		(PC.PlayerReplicationInfo.PlayerID == PlayerID) // Local Player Match
		&&
		(PC.Player != None)
		&&
		(PC.Player.Console != None)
	)
	{
		if (Multiplier > 1)
		{
			PC.Player.Console.Message
			(
				"*"
				@ GetCaptionString (AwardIndex)
				@ Class'ReUtility'.Static.GetColorCodeString (Default.AwardMultiplierColor)
				$ "x"
				$ Multiplier,
				0
			);
		}
		else
		{
			PC.Player.Console.Message
			(
				"*"
				@ GetCaptionString (AwardIndex)
				@ Class'ReUtility'.Static.GetColorCodeString (Default.AwardMultiplierColor),
				0
			);
		}
	}
}

// #

DefaultProperties
{
	CyanCaptionColor = (R=0,G=160,B=255,A=255)
	GrayCaptionColor = (R=23,G=204,B=23,A=255)
	OrangeCaptionColor = (R=224,G=128,B=16,A=255)
	PurpleCaptionColor = (R=152,G=81,B=204,A=255)
	RedCaptionColor = (R=255,G=0,B=0,A=255)
	WhiteCaptionColor = (R=223,G=223,B=223,A=255)
	CaptionColor = (R=223,G=223,B=223,A=255)
	AwardMultiplierColor = (R=223,G=223,B=31,A=255)
}
