// Developed 2011-2013 rejecht <rejecht at outlook dot com>

class LocalMessage_Pickup extends LocalMessage
	abstract
	;

var private const Material Pickup_Material_Default;

var private const IntBox
	Pickup_OffsetMaterial_Adrenaline,
	Pickup_OffsetMaterial_AssaultAmmunition,
	Pickup_OffsetMaterial_DoubleDamage,
	Pickup_OffsetMaterial_Health,
	Pickup_OffsetMaterial_Shield;

var(ConstantStrings) private localized const string
	AdrenalineString,
	DoubleDamageString,
	HealthVialString,
	HealthPackString,
	SuperHealthPackString,
	ShieldPackString,
	SuperShieldPackString;

var(ConstantStrings) protected const string
	ItemStringPrefix,
	ItemStringSuffix;

// #Queries

static event Color GetConsoleColor (PlayerReplicationInfo RelatedPRI_1)
{
	return Class'PlayerSettings'.Default.ItemColor;
}

static event Color GetColor
(
	optional int Data,
	optional PlayerReplicationInfo RelatedPRI_1,
	optional PlayerReplicationInfo RelatedPRI_2
)
{
	return Class'PlayerSettings'.Default.ItemColor;
}

// - Some stock pickups do not have an ItemName or IconCoords.

static function string GetShortPickupNameString (Class<Pickup> PickupClass)
{
	if (Class<TournamentHealth>(PickupClass) != None)
	{
		if (Class<MiniHealthPack>(PickupClass) != None)
		{
			return Default.HealthVialString;
		}
		else if (Class<SuperHealthPack>(PickupClass) != None)
		{
			return Default.SuperHealthPackString;
		}
		else if (Class<HealthPack>(PickupClass) != None)
		{
			return Default.HealthPackString;
		}
		else if
		(
			(PickupClass.Default.InventoryType != None)
			&&
			(PickupClass.Default.InventoryType.Default.ItemName != "")
		)
		{
			return PickupClass.Default.InventoryType.Default.ItemName;
		}
		else
		{
			return Default.HealthPackString;
		}
	}
	else if (Class<ShieldPickup>(PickupClass) != None)
	{
		if (Class<ShieldPack>(PickupClass) != None)
		{
			return Default.ShieldPackString;
		}
		else if (Class<SuperShieldPack>(PickupClass) != None)
		{
			return Default.SuperShieldPackString;
		}
		else
		{
			return Default.ShieldPackString;
		}
	}
	else if (Class<AdrenalinePickup>(PickupClass) != None)
	{
		return Default.AdrenalineString;
	}
	else if (Class<UDamagePack>(PickupClass) != None)
	{
		return Default.DoubleDamageString;
 	}
	else if
	(
		(PickupClass.Default.InventoryType != None)
		&&
		(PickupClass.Default.InventoryType.Default.ItemName != "")
	)
	{
		return PickupClass.Default.InventoryType.Default.ItemName;
	}
	else
	{
		return PickupClass.Static.GetLocalString ();
	}
}

// Preconditions:
// - (PickupClass != None)

protected static function bool GetIcon
(
	Class<Pickup> PickupClass,
	out Material Out_Material,
	out IntBox Out_OffsetMaterial
)
{
	if (Class<TournamentHealth>(PickupClass) != None)
	{
		Out_Material = Default.Pickup_Material_Default;
		Out_OffsetMaterial = Default.Pickup_OffsetMaterial_Health;

		return True;
	}
	else if (Class<ShieldPickup>(PickupClass) != None)
	{
		Out_Material = Default.Pickup_Material_Default;
		Out_OffsetMaterial = Default.Pickup_OffsetMaterial_Shield;

		return True;
	}
	else if (Class<AssaultAmmoPickup>(PickupClass) != None)
	{
		Out_Material = Default.Pickup_Material_Default;
		Out_OffsetMaterial = Default.Pickup_OffsetMaterial_AssaultAmmunition;

		return True;
	}
	else if (Class<AdrenalinePickup>(PickupClass) != None)
	{
		Out_Material = Default.Pickup_Material_Default;
		Out_OffsetMaterial = Default.Pickup_OffsetMaterial_Adrenaline;

		return True;
	}
	else if (Class<UDamagePack>(PickupClass) != None)
	{
		Out_Material = Default.Pickup_Material_Default;
		Out_OffsetMaterial = Default.Pickup_OffsetMaterial_DoubleDamage;

		return True;
	}
	else if (PickupClass.Default.InventoryType != None)
	{
		Out_Material = PickupClass.Default.InventoryType.Default.IconMaterial;
		Out_OffsetMaterial = PickupClass.Default.InventoryType.Default.IconCoords;

		return True;
	}
	else
	{
		Out_Material = None;

		return False;
	}
}

// #

DefaultProperties
{
	ItemStringPrefix = "["
	ItemStringSuffix = "]"

	AdrenalineString = "Adrenaline"
	DoubleDamageString = "Double Damage"
	HealthVialString = "Health Vial"
	HealthPackString = "Health Pack"
	SuperHealthPackString = "Big Keg O' Health"
	ShieldPackString = "Shield Pack"
	SuperShieldPackString = "Super Shield Pack"

	Pickup_Material_Default = Texture'HudContent.Generic.HUD'

	Pickup_OffsetMaterial_Adrenaline = (X1=113,Y1=38,X2=165,Y2=106)
	Pickup_OffsetMaterial_AssaultAmmunition = (X1=335,Y1=80,X2=382,Y2=126)
	Pickup_OffsetMaterial_DoubleDamage = (X1=0,Y1=164,X2=73,Y2=246)
	Pickup_OffsetMaterial_Health = (X1=74,Y1=165,X2=123,Y2=216)
	Pickup_OffsetMaterial_Shield = (X1=0,Y1=248,X2=68,Y2=310)

	// LocalMessage

	bFadeMessage = True
	bIsSpecial = False
	bIsUnique = True

	DrawColor = (R=255,G=255,B=255,A=255)
	FontSize = 0
	PosX = 0.5
	PosY = 0.92
	StackMode = SM_None
}
