// *F* Developed 2011-2013 rejecht <rejecht at outlook dot com>

class ReBot extends xBot
	HideDropdown
	;

// #PawnClass

// Contexts:
// - Server: GameInfo.Login ()

event SetPawnClass
(
	string BotPawnClassNameString,
	string BotCharacterNameString
)
{
	local Class<RePawn> LoadedPawnClass;

	LoadedPawnClass = Class<RePawn>(DynamicLoadObject (BotPawnClassNameString, Class'Core.Class'));

	if (LoadedPawnClass != None)
	{
		PawnClass = LoadedPawnClass;
	}

	PawnSetupRecord = Class'XGame.xUtil'.Static.FindPlayerRecord (BotCharacterNameString);

	PlayerReplicationInfo.SetCharacterName (BotCharacterNameString);
}

// ## Adrenaline

// - To prevent a bot from executing a disabled combo.
// - To conditionally disable adrenaline combos as specified for the
// server instance.

event TryCombo (string ComboClassNameString)
{
	if
	(
		bAdrenalineEnabled
		&&
		!ReGRI(Level.GRI).IsOver ()
		&&
		!Pawn.InCurrentCombo ()
		&&
		!NeedsAdrenaline ()
	)
	{
		ComboClassNameString = GetRecommendedComboNameString (ComboClassNameString);

		if (ComboClassNameString != "")
		{
			Pawn.DoComboName (ComboClassNameString);
		}
	}
}

// - Bots request a random adrenaline combo or a specific one.
// - The game or the mutator may recommend a different combo.
// - Any disabled combo overrides mutators (see game).

protected function string GetRecommendedComboNameString (string ComboClassNameString)
{
	if (ComboClassNameString ~= "Random")
	{
		ComboClassNameString = ComboNames[Rand (ArrayCount (ComboNames))];
	}
	else if (ComboClassNameString ~= "DMRandom")
	{
		ComboClassNameString = ComboNames[1 + Rand (ArrayCount (ComboNames) - 1)];
	}

	return Level.Game.NewRecommendCombo (ComboClassNameString, self);
}

// - RePlayer must have the same logic.

event AwardAdrenaline (float Amount)
{
	if (bAdrenalineEnabled)
	{
		if (Amount > 0.0f)
		{
			RePRI(PlayerReplicationInfo).Statistics_AdrenalineGained (Amount);
		}
		else if (Amount < 0.0f)
		{
			RePRI(PlayerReplicationInfo).Statistics_AdrenalineLost (Amount);
		}
	}


	Super.AwardAdrenaline (Amount);
}

// #Dead

state Dead
{
	event ExecuteWhatToDoNext () // *UDKFix*
	{
	}

	event WhatToDoNext (byte Context)
	{
	}

	event Timer ()
	{
		SetTimer (0.0f, False); // *UT2004Fix*
	}

	event bool IsDead ()
	{
		return True; // *UDKFix*
	}
}

event bool IsDead ()
{
	return False;
}

// #

state GameEnded
{
	Begin:
}

// #

state RoundEnded
{
	Begin:
}

// #Notifications

// - To skip the iteration through all players.
// - Note: Breaks ViewNextBot call.

event WasKilledBy (Controller Killer)
{
	if
	(
		(Killer != None)
		&&
		(Killer.Pawn != None)
	)
	{
		LastKillerPosition = Killer.Pawn.Location;
	}
}

// #Development

simulated event DisplayDebug (Canvas Canvas, out float YL, out float YPos)
{
	local float XL;

	Canvas.Font = Canvas.SmallFont;
	Canvas.StrLen ("*", XL, YL);
	YPos = YPos + YL;
	Canvas.SetPos (4, YPos);

	Super.DisplayDebug (Canvas, YL, YPos);
}

// #

DefaultProperties
{
	// Controller

	PawnClass = Class'RePawn' // Level.Game.GetDefaultPlayerClass ()
	PlayerReplicationInfoClass = Class'RePRI'
}
