// *F* Developed 2013 rejecht <rejecht at outlook dot com>

class ReFreon_Bot extends AMBot
	HideDropdown
	;

var(Constants) protected const Vector NullVector;

var(Cached) protected Pawn ThawingPawn;

// #Lifetime

// - When controllers are destroyed (balancing).

event Destroyed ()
{
	if (ThawingPawn != None)
	{
		ThawingPawn.Destroy ();
		ThawingPawn = None;
    }


	Super.Destroyed ();
}

// #PawnClass

event SetPawnClass
(
    string BotPawnClassNameString,
    string BotCharacterNameString
)
{
    local class<ReFreon_Pawn> LoadedPawnClass;

    LoadedPawnClass = Class<ReFreon_Pawn>(DynamicLoadObject (BotPawnClassNameString, Class'Core.Class'));

    if (LoadedPawnClass != None)
    {
        PawnClass = LoadedPawnClass;
    }

    PawnSetupRecord = Class'XGame.xUtil'.Static.FindPlayerRecord (BotCharacterNameString);

    PlayerReplicationInfo.SetCharacterName (BotCharacterNameString);
}

// #New

function bool IsThawing ()
{
	return False;
}

// #Adrenaline

// - Reminder: AMPlayer must have the same logic.
// - a) Adrenaline reduced by 20% in ReFreon.
// - Note that penalties are also reduced by 20%.

event AwardAdrenaline (float Amount)
{
	Super.AwardAdrenaline (Amount * 0.8f);
}

// #

function PawnFroze (Pawn P)
{
	if ( Pawn != P )
		return;


	if (Pawn != None)
	{
		ThawingPawn = Pawn;

		if (Pawn.InCurrentCombo ())
		{
			Adrenaline = 0;
        }


		SetLocation (Pawn.Location);


		if (Pawn.IsA ('ReFreon_Pawn'))
		{
			ReFreon_Pawn(Pawn).Frozen ();
        }
        else
        {
			Pawn.Unpossessed ();
        }
	}


	Pawn = None; // Replicated + Forces a reference count check?
	
	PendingMover = None;


	if (bIsPlayer)
    {
        if
		(
			!IsInState ('GameEnded')
			&&
			!IsInState ('RoundEnded')
		)
		{
			GotoState ('Thawing');
        }
    }
	else
	{
		Destroy ();
    }
}

// #

function PawnThawed (Pawn P)
{
}

// #

protected function FreezeCombatTimer ()
{
	SetTimer (0.0f, False);
}

// #

state Thawing
{
	ignores HitWall, Falling, TakeDamage, EnemyNotVisible;

	event HearNoise (float Loudness, Actor NoiseMaker)
	{
    }

	event bool AllowDetourTo (NavigationPoint N)
	{
		return False;
    }

	event SoakStop (string problem)
	{
    }

	event SetCombatTimer ()
	{
		SetTimer (0.0f, False);
    }

	event Pawn GetMyPlayer ()
	{
		return ReFreon_PRI(PlayerReplicationInfo).ThawingPawn;
    }

	event UpdatePawnViewPitch ()
	{
	    if (ReFreon_PRI(PlayerReplicationInfo).ThawingPawn != None)
	        ReFreon_PRI(PlayerReplicationInfo).ThawingPawn.SetViewPitch(Rotation.Pitch);
    }

	event ChangedWeapon ()
	{
    }

	event PawnDied (Pawn P)
	{
    }

    event bool CanImpactJump ()
    {
		return False;
    }

	event MayShootTarget ()
	{
    }

    event bool CanUseTranslocator()
    {
		return False;
    }

	event ImpactJump ()
	{
    }

    event WaitForMover (Mover M)
    {
    }

	event Trigger (Actor Other, Pawn EventInstigator)
	{
	}

	event MayDodgeToMoveTarget ()
	{
    }

    // Frequency: Tick

	event SeePlayer (Pawn SeenPlayer)
	{
    }

	event bool CanComboMoving ()
	{
		return False;
	}

	function bool CanCombo ()
	{
		return False;
    }

	function YellAt (Pawn Moron)
	{
    }

    event SetOrders (name NewOrders, Controller OrderGiver)
    {
    }

	event BeginState ()
	{
		bStasis = True;

		Enemy = None;

		FreezeCombatTimer ();

		PendingMover = None;
		Focus = None;

		FormerVehicle = None;
		bFrustrated = False;
		BlockedPath = None;
		bInitLifeMessage = False;
		bPlannedJump = false;
		bInDodgeMove = false;
		bReachedGatherPoint = false;
		bFinalStretch = false;
		bWasNearObjective = false;
		bPreparingMove = false;
		bEnemyEngaged = false;
		bPursuingFlag = false;
		bHasSuperWeapon = false;
		RouteGoal = None;
		MoveTarget = None;

		Pawn = None;
	}

	event EndState ()
	{
		bStasis = False;

		SetCombatTimer ();
    }

	event DelayedWarning ()
	{
	}

	event SwitchToBestWeapon ()
	{
	}

    event SetAttractionState ()
    {
    }

    function bool ClearShot (Vector TargetLoc, bool bImmediateFire)
    {
		return False;
    }

    function bool CanStakeOut ()
    {
		return False;
    }

    function CheckIfShouldCrouch (vector StartPosition, vector TargetPosition, float probability)
    {
    }

    function bool IsSniping()
    {
		return False;
    }

	event EnemyChanged (bool bNewEnemyVisible)
	{
    }

	event WanderOrCamp (bool bMayCrouch)
	{
    }

	function bool IsThawing ()
	{
		return True;
	}

	event bool IsDead ()
	{
		return True;
	}

	event CancelCampFor (Controller C)
	{
    }

	event float RelativeStrength (Pawn Other)
	{
		return 0.0f;
    }

    // Notify

	event bool NotifyPhysicsVolumeChange (PhysicsVolume PV)
	{
		local Pawn P;


		if (Pawn != None)
		{
			P = Pawn;
        }


        if (P != None)
        {
			if (Vehicle(P) != None)
			{
				return False;
	        }


	        if (PV.bWaterVolume)
			{
				bPlannedJump = False;


				if (!P.bCanSwim)
				{
					MoveTimer = -1.0;
                }
				else if (P.Physics != PHYS_Swimming)
				{
					P.setPhysics (PHYS_Swimming);
                }
			}
			else if (P.Physics == PHYS_Swimming)
			{
				if (P.bCanFly)
				{
					P.SetPhysics (PHYS_Flying);
                }
				else
				{
					P.SetPhysics (PHYS_Falling);
				}
			}
        }


		return False;
    }

    event bool NotifyHeadVolumeChange (PhysicsVolume PV)
    {
		return False;
    }

	event bool NotifyLanded (Vector HitNormal)
	{
		bNotifyFallingHitWall = False;

		bDodgingForward = False;
		bInDodgeMove = False;

		bPlannedJump = False;

		if (Pawn != None)
		{
			Pawn.Acceleration = NullVector;
        }

		return False;
    }

	event bool NotifyHitWall (Vector HitNormal, Actor Wall)
	{
		return False;
    }

	event bool NotifyBump (Actor Other)
	{
		Disable ('NotifyBump');

		return False;
    }

	//

    event NotifyPostLanded ()
    {
        bNotifyPostLanded = False;
    }

    event NotifyFallingHitWall (Vector HitNormal, Actor Wall)
    {
		bNotifyFallingHitWall = False;
    }

	//

    event InstantWarnTarget (Actor Target, FireProperties FiredAmmunition, vector FireDir)
    {
    }

	event ReceiveWarning (Pawn shooter, float projSpeed, vector FireDir)
	{
	}

	// - Keeps getting called because Pawn.Controller = Self.
	// Context:
	// - Projectile.PostBeginPlay ()

	event ReceiveProjectileWarning (Projectile NewProjectile)
	{
    }

    //

	event bool AdjustAround (Pawn P)
	{
		return True;
	}

	event MissedDodge ()
	{
    }

	event bool ShouldFireAgain (float RefireRate)
	{
		return False;
    }

	event TimedFireWeaponAtEnemy ()
	{
    }

    event bool FireWeaponAt (Actor A)
    {
		return False;
    }

    function bool CanAttack (Actor Other)
    {
		return False;
    }

	event bool WeaponFireAgain (float RefireRate, bool bFinishedFire)
	{
		bShootTarget = False;

		bShootSpray = False;

		StopFiring ();

		return False;
	}

	event Actor FaceActor (float StrafingModifier)
	{
		return None;
	}

	event ExecuteWhatToDoNext ()
	{
		Warn ("(Thawing)");
    }

	event bool DoWaitForLanding ()
	{
		Warn ("(Thawing)");

		return False;
    }

    function bool EnemyVisible ()
    {
		return False;
    }

    function DoRangedAttackOn(Actor A)
    {
    }

    function ChooseAttackMode ()
    {
    }

    function bool PickRetreatDestination ()
    {
		return False;
    }

    function bool FindRoamDest ()
    {
		return False;
    }

    function bool TestDirection (vector dir, out vector pick)
    {
		return False;
    }

	function bool CheckPathToGoalAround (Pawn P)
	{
		return False;
    }

	event ClearPathFor (Controller VIP)
	{
    }

	function DirectedWander (vector WanderDir)
	{
    }

    function bool PriorityObjective ()
    {
		return False;
    }

    function SetFall ()
    {
    }

    function bool StartMoveToward (Actor O)
    {
		return False;
    }

    function bool FindBestPathToward(Actor A, bool bCheckedReach, bool bAllowDetour)
    {
		return False;
    }

    function bool NeedToTurn (vector targ)
    {
		return False;
    }

    function bool NearWall (float walldist)
    {
		return False;
    }

    function bool CheckFutureSight(float deltatime)
    {
		return False;
    }

	function float AdjustAimError(float aimerror, float TargetDist, bool bDefendMelee, bool bInstantProj, bool bLeadTargetNow )
	{
		return 0.0f;
    }

    function bool TryWallDodge(vector HitNormal, actor HitActor)
    {
		return False;
    }

    function bool TryToDuck (vector duckDir, bool bReversed)
    {
		return False;
    }

    function Actor FaceMoveTarget ()
    {
		return None;
    }

    function bool ShouldStrafeTo (Actor WayPoint)
    {
		return False;
    }

    function Actor AlternateTranslocDest ()
    {
		return None;
    }

    function bool NeedWeapon ()
    {
		return False;
    }

    function DamageAttitudeTo(Pawn Other, float Damage)
    {
    }

    function bool IsRetreating ()
    {
		return False;
    }

    function bool Formation ()
    {
		return False; // ?
    }

    function Celebrate()
    {
    }

    function ForceGiveWeapon()
    {
    }

    function ForceCelebrate()
    {
    }

    function float GetDesiredOffset ()
    {
		return 0.0f;
    }

    function bool LostContact (float MaxTime)
    {
		return True;
    }

    function bool LoseEnemy ()
    {
		return True;
    }

	// Do

    function DoStakeOut()
	{
    }

    function DoCharge()
    {
    }

    function DoTacticalMove()
    {
    }

    function DoRetreat()
    {
    }

	//

	function bool DefendMelee (float Dist)
	{
		return False;
    }

	// Is

    function bool IsStrafing ()
    {
		return False;
    }

    function bool IsHunting ()
    {
		return False;
    }

	event bool IsShootingObjective ()
	{
		return False;
    }

	//

    function bool FindViewSpot ()
    {
		return False;
    }

    function bool Stopped ()
    {
		return True;
    }

	function SetEnemyReaction(int AlertnessLevel)
	{
    }

    function SetNewScript(ScriptedSequence NewScript)
    {
    }

    function bool ScriptingOverridesAI()
    {
		return False;
    }

	function bool ShouldPerformScript()
	{
		return False;
    }

	// AIController

	function int GetFacingDirection()
	{
		local float strafeMag;
		local vector Focus2D, Loc2D, Dest2D, Dir, LookDir, Y;

		// check for strafe or backup
		Focus2D = FocalPoint;
		Focus2D.Z = 0;
		Loc2D = Pawn.Location;
		Loc2D.Z = 0;
		Dest2D = Destination;
		Dest2D.Z = 0;
		lookDir = Normal(Focus2D - Loc2D);
		Dir = Normal(Dest2D - Loc2D);
		strafeMag = lookDir dot Dir;
		Y = (lookDir Cross vect(0,0,1));
		if ((Y Dot (Dest2D - Loc2D)) < 0)
			return ( 49152 + 16384 * strafeMag );
		else
			return ( 16384 - 16384 * strafeMag );
	}

	//

	event WhatToDoNext (byte Context)
	{
    }

    event Timer ()
    {
		FreezeCombatTimer ();
    }

	event Tick (float DeltaTime)
	{
		if (bPendingShoot)
		{
			bPendingShoot = False;
		}

		Disable ('Tick');
	}

	event ServerRestartPlayer ()
	{
    }

    //

    event float Desireability (Pickup P)
    {
		return 0.0f;
    }

    event float SuperDesireability (Pickup P)
    {
		return 0.0f;
    }

    event MonitoredPawnAlert ()
    {
    }

    event Weapon HasSuperWeapon ()
    {
		return None;
    }

    event bool ShouldKeepShielding ()
    {
		return False;
    }

    event GetOutOfVehicle ()
    {
    }

    event bool CanDoubleJump (Pawn Other)
    {
		return False;
    }

    event PickDestination ()
    {
    }

    event TryCombo (string ComboName)
    {
    }

    event FearThisSpot (AvoidMarker AvoidSpot)
    {
    }

    event Startle (Actor Feared)
    {
    }

	event LongFall ()
	{
    }

	event NotifyTakeHit (Pawn InstigatedBy, Vector HitLocation, int Damage, Class<DamageType> damageType, vector Momentum)
	{
    }

    event NotifyKilled (Controller Killer, Controller Killed, pawn KilledPawn)
    {
    }

    event NotifyHitMover (Vector HitNormal, Mover Wall)
    {
    }

    event NotifyJumpApex ()
    {
    }

    event NotifyMissedJump ()
    {
    }

	event bool SetRouteToGoal (Actor A)
	{
		return True;
	}

	event SetMoveTarget ()
	{
    }
}

state Dead
{
    event PawnDied (Pawn P)
	{
    }
}

state RestFormation
{
	event CancelCampFor (Controller C)
	{
		if
		(
			(Pawn != None)
			&&
			(C.Pawn != None)
		)
		{
			DirectedWander(Normal(Pawn.Location - C.Pawn.Location));
        }
	}
}

// #

DefaultProperties
{
	// Controller

	PawnClass = Class'ReFreon_Pawn'
	PlayerReplicationInfoClass = Class'ReFreon_PRI'
}
