// *F* Developed 2013-2014 rejecht <rejecht at outlook dot com>
// - Based in part on the implementation done by 3SPN Freon and L7 FreezeTag.
// - Note that this is a hack.
// - Unpossession is the correct way, but a controller is required to make pawn
// network relevant, and collide correctly with some actors and world.
// -- Using a dummy controller to possess the pawn just adds another layer of problems
// with controller, playerreplication, bIsPlayer, and so on.
// - Once frozen, the pawn becomes disposable, as it will be destroyed once thawed,
// and a new one gets spawned.	Reusing the pawn had inventory replication issues.

class ReFreon_Pawn extends AMPawn
	config
	HideDropdown
	;

var(Game) protected ReFreon_Aura Aura;

var array<ReTAM.SLoadout> SavedLoadout;

var protected const Class<DamageType> ShatteredDamageTypeClass;

var(Game) float LastThawedTeammateTime;

var(Gib) protected const Class<xPawnGibGroup> FrozenGibGroupClass;

var(ConstantSounds) protected const Sound
	LandingHitSound[6],
	IceHitSound,
	LavaThawSound,
	MoverHitSound,
	ThawingAmbientSound;

var protected float LastMoverHitSoundTimeSeconds;

var(Replicated) protected bool bThawing;
var(Proxy) protected bool bProxyThawing;

var(Glocals) protected Rotator FaceRotation_L_NewRotation;

var(Game) bool bWasThawed;

// #Replication

replication
{
	reliable if (Role == ROLE_Authority)
		ClientFroze
		;

	reliable if (Role == ROLE_Authority)
		bThawing
		;
}

// #Lifetime

simulated event Destroyed ()
{
	if (Level.NetMode != NM_Client)
	{
		Aura_Destroyed ();
	}


	Super.Destroyed ();
}

// #Possession

event PossessedBy (Controller Other)
{
	Super.PossessedBy (Other);


	if (Level.NetMode != NM_Client)
	{
		Aura_PostBeginPlay ();
	}
}

// #Weapons

// - Called on GameInfo.DiscardInventory to store weapon status.

function StoreWeaponStatus (Class<Weapon> WeaponClass, int Primary, int Secondary)
{
	local int Index;

	Index = SavedLoadout.Length;
	SavedLoadout.Length = Index + 1;

	SavedLoadout[Index].WeaponClass = WeaponClass;
	SavedLoadout[Index].Primary = Primary;
	SavedLoadout[Index].Secondary = Secondary;

	if (SavedLoadout.Length > 16)
	{
		Warn ("(SavedLoadout.Length > 16)");
	}
}

// #

event Died
(
	Controller Killer,
	Class<DamageType> DamageTypeClass,
	Vector HitLocation
)
{
	if
	(
		!bDeleteMe
		&&
		!Level.bLevelChange
		&&
		(Level.Game != None)
	)
	{
		if (DamageTypeClass == None)
		{
			DamageTypeClass = Class'Engine.DamageType'; // Implies (!bCausedByWorld)
		}
		else if
		(
			DamageTypeClass.Default.bCausedByWorld
			&&
			(
				(Killer == None)
				||
				(Killer == Controller)
			)
			&&
			(LastHitBy != None)
		)
		{
			Killer = LastHitBy;
		}

		// Determine whether pawn should freeze or die.

		if
		(
			!Level.Game.PreventDeath (Self, Killer, DamageTypeClass, HitLocation)
			&&
			ReFreon(Level.Game).Static.FreezingDamageType (DamageTypeClass)
			&&
			!IsInPain () // Ignore freezing a pawn on weapon damage when located in a damaging environment (will just get thawed again immediately, in which case would get average health, and thaw protection).
		)
		{
			Froze (Killer, DamageTypeClass);
		}
		else
		{
			Super.Died (Killer, DamageTypeClass, HitLocation);
		}
	}
}

// - The state of the player is about to be set server-side.

function Frozen ()
{
	local PlayerController PC;


	Acceleration = NullVector;


	bReplicateAnimations = False;


	NetUpdateTime = Level.TimeSeconds - 1;

	if (DrivenVehicle != None)
	{
		NetUpdateFrequency = 5;
	}


	PC = PlayerController(Controller);

	if (PC != None)
	{
		ClientFroze (PC);
	}
}

// #

simulated function ClientFroze (PlayerController PC)
{
	Acceleration = NullVector;

	PC.GotoState ('PlayerThawing');
}

// Extended to remove all emitters attached to bones.

simulated event RemoveFlamingEffects ()
{
	local Emitter E;
	local int Index;
	local int IndexEmitter;
	local int Count;

	Count = Attached.Length;

	for (Index = 0; Index < Count; ++Index)
	{
		if (Attached[Index] != None)
		{
			if (Attached[Index].IsA ('xEmitter'))
			{
				xEmitter(Attached[Index]).mRegen = False;
			}
			else if (Attached[Index].IsA ('Emitter'))
			{
				E = Emitter(Attached[Index]);

				for (IndexEmitter = 0; IndexEmitter < E.Emitters.Length; ++IndexEmitter)
				{
					E.Emitters[IndexEmitter].SkeletalMeshActor = None;
				}

				E.Kill ();
			}
		}
	}
}

// 1. Died ()
// 2. Froze ()

protected function Froze (Controller Killer, Class<DamageType> DamageTypeClass)
{
	bThawing = True;

	ReFreon_PRI(PlayerReplicationInfo).ThawingPawn = Self;

	Health = 1; // At least 1 because of RedeemerWarhead, and technically because you are not dead, though you are considered out.

	// Do not throw held weapon.

	if
	(
		(Weapon != None)
		&&
		(
			(DrivenVehicle == None)
			||
			DrivenVehicle.bAllowWeaponToss
		)
		&&
		(Controller != None)
	)
	{
		Controller.LastPawnWeapon = Weapon.Class; // Was supposed to use this to restore the last held weapon.	Was originally used by statistics to show what weapon the player held on death.


		Weapon.HolderDied (); // To stop weapon from firing and force a release.
	}

	// Vehicles are not supported, but anyway..

	if (DrivenVehicle != None)
	{
		Velocity = DrivenVehicle.Velocity; // Character inherits velocity of vehicle.


		DrivenVehicle.DriverDied ();
	}

	// Scoring
	// - Inventory gets discarded, however, current ammo status gets recorded into this pawn for when it thaws.
	// -- Keeping inventory seemed to be too complicated, and leads to native crashes server-side.
	// - bOutofLives gets set since player has one life and is considered out when thawing.

	if (Controller != None)
	{
		Level.Game.Killed (Killer, Controller, Self, DamageTypeClass);
	}
	else
	{
		Level.Game.Killed (Killer, Controller(Owner), Self, DamageTypeClass);
	}

	// Powerups

	RemovePowerups ();

	// Vehicle
	// - Delayed setting None because referenced by Killed.

	DrivenVehicle = None;


	TearoffMomentum += Velocity;
	Velocity = NullVector;
	AddVelocity (TearoffMomentum);
	TearoffMomentum = NullVector; // Reset as to not kick in on ChunkUp and later PlayDying calls.


	PlayFrozen ();


	if (Controller != None)
	{
		if (Controller.bIsPlayer)
		{
			if (Controller.IsA ('ReFreon_Player'))
			{
				ReFreon_Player(Controller).PawnFroze (Self);
			}
			else if (Controller.IsA ('ReFreon_Bot'))
			{
				ReFreon_Bot(Controller).PawnFroze (Self);
			}
			else
			{
				Warn ("!ReFreon_Bot && !ReFreon_Player");
			}
		}
		else
		{
			Warn ("Controller.Destroy ()");

			Controller.Destroy ();
		}
	}
}

// - Called on proxies in netplay based on bThawing.

protected simulated function PlayFrozen ()
{
	ResetHitFX (); // All replicated hit effects are reset (such as flames, dripping blood, ...).

	HitDamageType = ShatteredDamageTypeClass; // All pawns see this damage type.

	GibGroupClass = FrozenGibGroupClass; // Replace with general gibs.

	Acceleration = NullVector;

	SetPhysics (PHYS_Falling);

	if (Level.NetMode != NM_DedicatedServer)
	{
		RemoveFlamingEffects ();


		bScriptPostRender = True; // Custom Beacons
	}


	SetOverlayMaterial (None, 0.0f, True); // Replicated and overridden client-side.


	GotoState ('Thawing');


	if (Level.NetMode != NM_DedicatedServer)
	{
		ClientSetup ();
	}
}

simulated event TickFX (float DeltaTime)
{
	if
	(
		(Level.NetMode == NM_Client)
		&&
		(bThawing ^^ bProxyThawing) // The rare use of an exclusive or.
	)
	{
		bProxyThawing = bThawing;

		PlayFrozen ();
	}


	Super.TickFX (DeltaTime);
}

// #Aura

protected function Aura_PostBeginPlay ()
{
	if (Aura == None)
	{
		Aura = Spawn (ReFreon(Level.Game).GetAuraClass (), Self, , Location, Rotation);


		if (Aura == None)
		{
			Warn ("(Aura == None) - AuraClassName <" $ ReFreon(Level.Game).GetAuraClassName () $ ">");
		}
	}
}

protected function Aura_Destroyed ()
{
	if (Aura != None)
	{
		Aura.Destroy ();
		Aura = None;
	}
}

protected function StartThawingAura ()
{
	Aura.GotoState ('Thawing');
}

// #Queries

simulated function bool IsThawing ()
{
	return bThawing;
}

function bool WasThawed ()
{
	return bWasThawed;
}

// #Possession

// - Dumb workaround for Unpossessed setting PRI to None at end of game or on dying pawn.
// - PRI currently needed to look up materials and team at end of game.
// -- Later: Move frozen material to pawn DefaultProperties, as it is part of the mod.

event Unpossessed ()
{
	local PlayerReplicationInfo CachedPlayerReplicationInfo;

	CachedPlayerReplicationInfo = PlayerReplicationInfo;

	Super.Unpossessed ();

	PlayerReplicationInfo = CachedPlayerReplicationInfo;
}

// #

state Thawing
{
	// Pawn enters the thawing state.

	simulated event BeginState ()
	{
		AmbientSound = ThawingAmbientSound;
		
		SoundRadius = 40.0f;

		if (Level.NetMode != NM_Client)
		{
			bCanPickupInventory = False;


			bProjTarget = False;


			StartThawingAura ();


			bCanClimbLadders = False;
			bCanDoubleJump = False;
			bCanFly = False;
			bCanJump = False; // AI
			bCanStrafe = False;
			bCanSwim = False;
			bCanUse = False;
			bCanWalk = False;
			bCanWallDodge = False;
		}

		// Animation

		bPhysicsAnimUpdate = False;

		StopAnimating (True);

		// Permissions

		bNoWeaponFiring = True;

		//

		Weapon = None;
		PendingWeapon = None;
	}

	// Context:
	// - Self.Destroyed ()
	// - Self.LeaveThawingState ()

	simulated event EndState ()
	{
		AmbientSound = None; // Replicated

		SoundPitch = Default.SoundPitch;
		SoundRadius = Default.SoundRadius;
		SoundVolume = Default.SoundVolume;


		if (Level.NetMode != NM_DedicatedServer)
		{
			bScriptPostRender = False;
		}


		if (Level.NetMode != NM_Client)
		{
			Aura_Destroyed ();

			if (ReFreon_PRI(PlayerReplicationInfo) != None)
			{
				ReFreon_PRI(PlayerReplicationInfo).ThawingPawn = None;
			}
		}
	}

	// Called by thawing mechanism and the pawn.

	function Thaw (name Context, int Health)
	{
		if
		(
			!bDeleteMe
			&&
			!Level.bLevelChange
			&&
			(Level.Game != None)
		)
		{
			if (ReGRI(Level.GRI).IsLive ())
			{
				if (Context == 'Pain')
				{
					Shattered ();
				}
				else
				{
					Thawed (Context, Health);
				}
			}
			else
			{
				Shattered ();
			}
		}
	}

	protected function Thawed (name Cause, int Health)
	{
		ReFreon(Level.Game).Thawed (Self, Health, Cause); // Destroys this pawn which leaves this state.
	}

	// - Removed TakeDamage calls.

	simulated function bool IsThawing ()
	{
		return True;
	}

	// #

	exec event NextItem ()
	{
		Warn ("*");
	}

	exec event SwitchToLastWeapon ()
	{
		Warn ("*");
	}

	simulated event PrevWeapon ()
	{
		Warn ("*");
	}

	simulated event NextWeapon ()
	{
		Warn ("*");
	}

	event TakeFallingDamage ()
	{
		if (Velocity.Z < (-0.5f * MaxFallSpeed))
		{
			if (Role == ROLE_Authority)
			{
				MakeNoise (1.0f);
			}

			if (Controller != None)
			{
				Controller.DamageShake
				(
					FMin (1.0f, -1.0f * Velocity.Z / MaxFallSpeed)
				);
			}
		}
		else if (Velocity.Z < (-1.4f * JumpZ))
		{
			if (Role == ROLE_Authority)
			{
				MakeNoise (0.5f);
			}
		}
		else if (Velocity.Z < (-0.8f * JumpZ))
		{
			if (Role == ROLE_Authority)
			{
				MakeNoise (0.2f); // *UDKFix* To add less noise for shorter falls.
			}
		}
	}

	// Do not avoid death.

	event FellOutOfWorld (EKillZType KillType)
	{
		if (Level.NetMode != NM_Client)
		{
			Health = -1;

			if (KillType == KILLZ_Lava)
			{
				Super.Died (None, Class'Engine.FellLava', Location);
			}
			else if (KillType == KILLZ_Suicide)
			{
				Super.Died (None, Class'Engine.Fell', Location);
			}
			else
			{
				if (Physics != PHYS_Karma)
				{
					SetPhysics (PHYS_None);
				}

				Super.Died (None, Class'Engine.Fell', Location);
			}
		}
	}

	// Ignore PlayLanded.

	event Landed (Vector HitNormal)
	{
		local Rotator NewRotation;


		ImpactVelocity = NullVector;


		TakeFallingDamage ();


		if
		(
			(Velocity.Z < -200.0f)
			&&
			(PlayerController(Controller) != None)
		)
		{
			bJustLanded = PlayerController(Controller).bLandingShake;


			OldZ = Location.Z;
		}


		LastHitBy = None;


		PlaySound (Default.LandingHitSound[Rand (6)], SLOT_Pain, 1.5f * TransientSoundVolume);


		Velocity = NullVector;


		NewRotation = Rotation; // Fixed Pawn Rotation

		NewRotation.Pitch = 0;
		NewRotation.Roll = 0;

		SetRotation (NewRotation);
	}

	event HitWall (Vector HitNormal, Actor Wall)
	{
	}

	protected simulated function PlayFrozen ()
	{
		Warn ("(Thawing)");
	}

	simulated event PlayNextAnimation ()
	{
	}

	// Ingore changing the controller state.

	event DropToGround ()
	{
		bCollideWorld = True;

		bInterpolating = False;

		SetCollision (True, True);

		SetPhysics (PHYS_Falling);

		AmbientSound = None;
	}


	event TakeDamage
	(
		int Damage,
		Pawn PawnInstigator,
		Vector HitLocation,
		Vector HitMomentum,
		Class<DamageType> DamageTypeClass
	)
	{
		if (Role == ROLE_Authority)
		{
			if (DamageTypeClass == None)
			{
				DamageTypeClass = Class'Engine.DamageType';
			}


			LastHitBy = None;


			if (ReFreon(Level.Game).Static.FreezingDamageType (DamageTypeClass))
			{
				if
				(
					(Physics == PHYS_None)
					&&
					(DrivenVehicle == None)
				)
				{
					SetMovementPhysics ();
				}


				if
				(
					(Physics == PHYS_Walking)
					&&
					DamageTypeClass.Default.bExtraMomentumZ
				)
				{
					HitMomentum.Z = FMax (HitMomentum.Z, 0.4f * VSize (HitMomentum));
				}


				HitMomentum /= (Mass * 1.5f);


				PlayHit (Damage, PawnInstigator, HitLocation, DamageTypeClass, HitMomentum);


				if (HitLocation == NullVector)
				{
					HitLocation = Location;
				}


				AddVelocity (HitMomentum);
			}
			else if (ReFreon(Level.Game).ThawingDamageType (DamageTypeClass))
			{
				if (Class<FellLava> (DamageTypeClass) != None)
				{
					PlayOwnedSound
					(
						LavaThawSound,
						SLOT_Misc
					);

					Spawn (Class'PickupEffect',,, Location, Rotator (Vect (0.0f, 0.0f, 1.0f)));
				}


				Thaw ('ThawingDamageType', 0);
			}
			else if (IsInPain ())
			{
				Shattered ();
			}
			else if (DamageTypeClass.Default.bCausedByWorld)
			{
				Shattered ();
			}
		}
		else
		{
			if (Role < ROLE_Authority)
			{
				Warn ("(Role < ROLE_Authority)" @ DamageTypeClass);
			}


			if (Health < 0)
			{
				Warn ("(Health < 0)" @ DamageTypeClass);
			}


			if
			(
				(DamageTypeClass == None)
				&&
				(PawnInstigator != None)
			)
			{
				Warn ("(DamageTypeClass == None) && (PawnInstigator != None)" @ PawnInstigator);
			}
		}
	}

	// - Always spawn giblets (ignores LowGore preference).

	simulated event SpawnGiblet
	(
		Class<Gib> GibClass,
		Vector Location,
		Rotator Rotation,
		float GibPerterbation
	)
	{
		local Gib Giblet;
		local Vector Direction, Dummy;

		if (GibClass != None)
		{
			Instigator = Self;

			Giblet = Spawn (GibClass,,, Location, Rotation);

			if (Giblet != None)
			{
				Giblet.bFlaming = bFlaming;
				Giblet.SpawnTrail ();

				GibPerterbation *= 32768.0;

				Rotation.Pitch += ( FRand() * 2.0 * GibPerterbation ) - GibPerterbation;
				Rotation.Yaw += ( FRand() * 2.0 * GibPerterbation ) - GibPerterbation;
				Rotation.Roll += ( FRand() * 2.0 * GibPerterbation ) - GibPerterbation;
			
				GetAxes( Rotation, Dummy, Dummy, Direction );
			
				Giblet.Velocity = Velocity + Normal(Direction) * (250 + 260 * FRand());
				Giblet.LifeSpan = Giblet.LifeSpan + 2 * FRand() - 1;
			}
		}
	}

	event HeadVolumeChange (PhysicsVolume newHeadVolume)
	{
	}

	event BreathTimer ()
	{
	}

	event Falling ()
	{
	}

	simulated event ChangeAnimation ()
	{
	}

	simulated event AnimEnd (int Channel)
	{
	}

	simulated event StopPlayFiring ()
	{
		bSteadyFiring = False;
	}

	// #Died

	event Died
	(
		Controller Killer,
		Class<DamageType> DamageTypeClass,
		Vector HitLocation
	)
	{
		Health = 0;

		Super.Died (None, ShatteredDamageTypeClass, HitLocation); // Ignore Global.Died () because already thawing.
	}


	simulated event PlayDying (Class<DamageType> DamageTypeClass, Vector HitLoc)
	{
		PlayShattered ();
	}


	protected function Shattered ()
	{
		Health = 0;

		Super.Died (None, ShatteredDamageTypeClass, Location);
	}

	event KilledBy (Pawn IgnoredEventInstigator)
	{
		Health = 0;

		Super.Died (None, ShatteredDamageTypeClass, Location);
	}

	event GibbedBy (Actor Other)
	{
		if (Role == ROLE_Authority)
		{
			if (Pawn(Other) != None)
			{
				if
				(
					(Pawn(Other).Weapon != None)
					&&
					Pawn(Other).Weapon.IsA ('Translauncher')
				)
				{
					Health = 0;

					Super.Died (None, Pawn(Other).Weapon.GetDamageType (), Location);
				}
				else
				{
					Health = 0;

					Super.Died (None, ShatteredDamageTypeClass, Location);
				}
			}
			else
			{
				Health = 0;

				Super.Died (None, ShatteredDamageTypeClass, Location);
			}
		}
	}

	function Actor GetPathTo (Actor Dest)
	{
		return None;
	}

	event ChooseFireAt (Actor A)
	{
	}

	simulated function Fire (optional float F )
	{
	}

	simulated function AltFire( optional float F )
	{
	}

	function bool RecommendLongRangedAttack ()
	{

		return False;
	}

	function bool CanAttack(Actor Other)
	{
		return False;
	}

	function bool TooCloseToAttack (Actor Other)
	{
		return False;
	}

	function float RefireRate()
	{
		return 0.0f;
	}

	function bool IsFiring ()
	{
		if (Weapon != None)
			return Weapon.IsFiring();

		return false;
	}

	function bool FireOnRelease()
	{
		if (Weapon != None)
			return Weapon.FireOnRelease();

		return false;
	}

	function bool NeedToTurn (vector Targ)
	{
		return False;
	}

	function float ModifyThreat (float current, Pawn Threat)
	{
		return Current;
	}

	function bool PerformDodge (eDoubleClickDir DoubleClickMove, vector Dir, vector Cross)
	{
		return False;
	}

	function bool CanGrabLadder()
	{
		return False;
	}

	function EndClimbLadder(LadderVolume OldLadder)
	{
	}

	function ClimbLadder (LadderVolume L)
	{
	}

	// Call-Frequency: Tick

	function bool NearMoveTarget()
	{
		return False;
	}

	function Actor GetMoveTarget()
	{
		return None;
	}

	function SetMoveTarget(Actor NewTarget )
	{
	}

	function bool LineOfSightTo (actor Other)
	{
		return False;
	}

	function Actor ShootSpecial (Actor A)
	{
		return None;
	}

	function float AdjustedStrength()
	{
		return 0;
	}

	function HandlePickup(Pickup pick)
	{
	}

	function Trigger( actor Other, pawn EventInstigator )
	{
	}

	function bool CanTrigger (Trigger T)
	{
		return False;
	}

	event FinishedInterpolation ()
	{
		DropToGround ();
	}

	function JumpOutOfWater(vector jumpDir)
	{
	}

	event ShouldCrouch (bool bShouldCrouch)
	{
	}

	// (Frequency) High
	// - The rotation is locked while the pawn is thawing.
	// - Keep a flat horizontal view.

	simulated event FaceRotation (Rotator IgnoredRotation, float DeltaTime)
	{
		FaceRotation_L_NewRotation = Rotation; // Fixed Pawn Rotation

		FaceRotation_L_NewRotation.Pitch = 0;
		FaceRotation_L_NewRotation.Roll = 0;

		SetRotation (FaceRotation_L_NewRotation);
	}

	// #Adrenaline

	event DoCombo (Class<Combo> ComboClass)
	{
	}

	event DoComboName (string ComboClassName)
	{
	}

	event bool InCurrentCombo ()
	{
		return False;
	}

	//

	event int CanUseShield (int Amount)
	{
		return Amount;
	}

	// #Weapon

	simulated event bool CanThrowWeapon ()
	{
		return False;
	}

	event TossWeapon (Vector TossVelocity)
	{
	}

	event bool IsHeadShot (Vector HitLocation, Vector Ray, float AdditionalScale)
	{
		return False;
	}

	event JumpOffPawn ()
	{
		Velocity += (100.0f + CollisionRadius) * VRand ();
		Velocity.Z = 200.0f + CollisionHeight;

		SetPhysics (PHYS_Falling);
	}

	function Gasp()
	{
	}

	function int ShieldAbsorb( int damage )
	{
		return damage;
	}

	event TakeDrowningDamage ()
	{
	}

	function bool CheckWaterJump (out vector WallNormal)
	{
		return False;
	}

	event DoDoubleJump (bool bUpdating)
	{
	}

	event bool CanDoubleJump()
	{
		return False;
	}

	event bool CanMultiJump ()
	{
		return False;
	}

	function bool Dodge (EDoubleClickDir DoubleClickMove)
	{
		return False;
	}

	function bool DoJump( bool bUpdating )
	{
		return False;
	}

	event PlayMoverHitSound ()
	{
		if ((Level.TimeSeconds - LastMoverHitSoundTimeSeconds) > 0.1f)
		{
			PlaySound
			(
				MoverHitSound,
				SLOT_Interact,
				0.5f, // Volume
				,
				,
				1.5f // Pitch
			);

			LastMoverHitSoundTimeSeconds = Level.TimeSeconds;
		}
	}

	event PlayDyingSound ()
	{
	}

	event PlayHit (float Damage, Pawn InstigatedBy, Vector HitLocation, Class<DamageType> DamageTypeClass, Vector Momentum)
	{
		if
		(
			!bDeleteMe
			&&
			!Level.bLevelChange
			&&
			(Level.Game != None)
		)
		{
			if
			(
				!ReGRI(Level.GRI).IsLive ()
				&&
				(
					(
						(Class<DamTypeShieldImpact>(DamageTypeClass) != None) // Minor shield impact pushes player around.
						&&
						(Damage <= 20)
					)
					||
					(Class<DamTypeShieldImpact>(DamageTypeClass) == None) // Other damage
				)
			)
			{
				Shattered ();
			}
			else if ((Level.TimeSeconds - LastPainTime) > 0.1f)
			{
				PlayTakeHit (HitLocation, Damage, DamageTypeClass);


				LastPainTime = Level.TimeSeconds;
			}
		}
	}

	event PlayTakeHit (Vector HitLocation, int Damage, Class<DamageType> DamageTypeClass)
	{
		if (Damage > 0)
		{
			PlayOwnedSound
			(
				IceHitSound,
				SLOT_Misc,
				0.5f, // Volume
				,
				,
				1.5f // Pitch
			);
		}
	}

	function bool IsInLoadout(class<Inventory> InventoryClass)
	{
		return true;
	}

	simulated event PlayFiring (optional float Rate, optional name FiringMode)
	{
	}

	function bool CannotJumpNow()
	{
		return True;
	}

	simulated event PlayJump()
	{
	}

	simulated event PlayFalling()
	{
	}

	simulated function PlayMoving()
	{
	}

	simulated event PlayWaiting ()
	{
	}

	function PlayLanded (float impactVel)
	{
	}

	simulated event PlayLandingAnimation (float ImpactVel)
	{
	}

	function PlayVictoryAnimation()
	{
	}

	function float RangedAttackTime()
	{
		return 0;
	}

	simulated event Vector GetTargetLocation () // OK
	{
		return Location;
	}

	simulated function bool FindValidTaunt( out name Sequence )
	{
		return False;
	}

	// UnrealPawn

	function EndJump()
	{
	}

	simulated function PlayFootStep (int Side)
	{
	}

	simulated function ShouldUnCrouch ()
	{
	}

	function FootStepping(int side)
	{
	}

	//

	simulated event AttachEffect (Class<xEmitter> EmitterClass, Name BoneName, Vector Location, Rotator Rotation)
	{
	}

	// xPawn

	simulated event StartFiring (bool bHeavy, bool bRapid)
	{
	}

	simulated event StopFiring ()
	{
		if (FireState == FS_Looping)
		{
			FireState = FS_PlayOnce;
		}


		IdleTime = Level.TimeSeconds;
	}

	simulated function PlayFootStepLeft()
	{
	}

	simulated function PlayFootStepRight()
	{
	}

	//

	simulated event bool IsPlayerPawn ()
	{
		return False; // To avoid triggering movers.
	}

	// #

	protected simulated function ClientSetup ()
	{
		local ReFreon_PRI OwnerPRI;
		local int IndexTeam_Owner;
		local int IndexSkin;
		local int Count;

		OwnerPRI = ReFreon_PRI(PlayerReplicationInfo);
		IndexTeam_Owner = GetTeamNum ();

		if
		(
			(OwnerPRI != None)
			&&
			(OwnerPRI.ClientSkins_CombinerSkin_FrozenBody != None)
			&&
			(IndexTeam_Owner < IndexTeam_Max)
			&&
			(ReFreon_GRI(Level.GRI) != None)
		)
		{
			if (ClientSkins_CustomColors ())
			{
				OwnerPRI.ClientSkins_CombinerSkin_FrozenBody.Material1 = ReGRI(Level.GRI).ClientSkins_ConstantColor_Team[IndexTeam_Owner];

				Count = Clamp (Skins.Length, 2, 4);

				for (IndexSkin = 0; IndexSkin < Count; ++IndexSkin)
				{
					Skins[IndexSkin] = OwnerPRI.ClientSkins_CombinerSkin_FrozenBody;
				}
			}
			else
			{
				if (ClientSkins_CachedMaterial_Skin[0] != None)
				{
					OwnerPRI.ClientSkins_CombinerSkin_FrozenBody.Material1 = ClientSkins_CachedMaterial_Skin[0];
				}
				else
				{
					OwnerPRI.ClientSkins_CombinerSkin_FrozenBody.Material1 = Default.Skins[0];
				}

				if (ClientSkins_CachedMaterial_Skin[1] != None)
				{
					OwnerPRI.ClientSkins_CombinerSkin_FrozenHead.Material1 = ClientSkins_CachedMaterial_Skin[1];
				}
				else
				{
					OwnerPRI.ClientSkins_CombinerSkin_FrozenHead.Material1 = Default.Skins[1];
				}

				Skins[0] = OwnerPRI.ClientSkins_CombinerSkin_FrozenBody;
				Skins[1] = OwnerPRI.ClientSkins_CombinerSkin_FrozenHead;
			}

			Count = Clamp (Skins.Length, 2, 4);

			for (IndexSkin = 0; IndexSkin < Count; ++IndexSkin)
			{
				RealSkins[IndexSkin] = Skins[IndexSkin];
			}
		}
	}
}

// #

function Thaw (name Cause, int Health)
{
	Warn ("(Global)");
}

protected function Thawed (int Health, name Cause)
{
	Warn ("(Global)");
}

// #

protected simulated function PlayShattered ()
{
	local class<Gib> CachedGibClass;
	local Vector HitNormal;
	local int Index;

	HitDamageType = ShatteredDamageTypeClass;

	TakeHitLocation = NullVector;

	AmbientSound = None;

	bCanTeleport = False;
	bReplicateMovement = False;
	bSkeletized = True; // Cheap way to get around playing death scream.
	bTearOff = True; // Calls PlayDying on client.
	bPlayedDeath = True;

	LifeSpan = RagdollLifeSpan;

	if (Level.NetMode != NM_DedicatedServer)
	{
		CachedGibClass = FrozenGibGroupClass.Static.GetGibClass (EGT_Head);

		for (Index = 0; Index < 8; ++Index)
		{
			HitNormal = Normal (Vect (0.0f, 0.0f, 1.0f) + VRand () * 0.2f + Vect (0.0f, 0.0f, 2.8f));

			SpawnGiblet (CachedGibClass, Location, Rotator(HitNormal), 1.0f);
		}
	}

	GotoState ('Dying');


	BaseEyeHeight = Default.BaseEyeHeight;


	SetInvisibility (0.0f);


	if (Physics == PHYS_Walking)
	{
		Velocity = NullVector;
	}


	bHidden = True;
}

// #HitFX

// - DamType is checked by ProcessHitFX client-side.

protected simulated function ResetHitFX ()
{
	HitFX[0].DamType = None;
	HitFX[1].DamType = None;
	HitFX[2].DamType = None;
	HitFX[3].DamType = None;
	HitFX[4].DamType = None;
	HitFX[5].DamType = None;
	HitFX[6].DamType = None;
	HitFX[7].DamType = None;
}

// #

protected function Shattered ()
{
	Warn ("(Global)");
}

// #

// - Used when thawing player when thaw protection is active.

function PlayThawedEffect ()
{
	if
	(
		!bSpawnIn
		&&
		(
			(Level.TimeSeconds - SpawnTime)
			<
			DeathMatch(Level.Game).SpawnProtectionTime
		)
	)
	{
		bSpawnIn = True;

		SetSpawnProtectionMaterial ();
	}
}

simulated function DisplayDebug(Canvas Canvas, out float YL, out float YPos)
{
	Super.DisplayDebug (Canvas, YL, YPos);

	//

	YPos += YL;
	Canvas.SetPos (4, YPos);
	Canvas.SetDrawColor (0, 255, 255);

	Canvas.DrawText ("bThawing:" @ bThawing @ "bProxyThawing:" @ bProxyThawing, False);
}

// #

DefaultProperties
{
	IceHitSound = Sound'WeaponSounds.BaseGunTech.BGrenfloor1'
	MoverHitSound = Sound'WeaponSounds.BaseGunTech.BGrenfloor1'
	LavaThawSound = Sound'GeneralAmbience.steamfx4'
	ThawingAmbientSound = Sound'Re.Sounds.Thawing'

	FrozenGibGroupClass = Class'xFrozenGibGroup'

	ShatteredDamageTypeClass = Class'DamTypeShattered'

	LandingHitSound(0) = Sound'PlayerSounds.BFootsteps.BFootstepSnow1'
	LandingHitSound(1) = Sound'PlayerSounds.BFootsteps.BFootstepSnow2'
	LandingHitSound(2) = Sound'PlayerSounds.BFootsteps.BFootstepSnow3'
	LandingHitSound(3) = Sound'PlayerSounds.BFootsteps.BFootstepSnow4'
	LandingHitSound(4) = Sound'PlayerSounds.BFootsteps.BFootstepSnow5'
	LandingHitSound(5) = Sound'PlayerSounds.BFootsteps.BFootstepSnow6'

	// Pawn

	ControllerClass = Class'ReFreon_Bot'
}
