// *F* Developed 2011-2013 rejecht <rejecht at outlook dot com>

class ReTDM_Scoreboard extends ReTeamGameScoreboard
	config (User)
	HideDropDown
	;

const IndexTeam_Max = 2;
const IndexTeam_A = 0;
const IndexTeam_B = 1;
const IndexTeam_Unspecified = 255;

const MinSlotsDrawn = 1;
const MaxSlotsDrawn = 7;

var(Tick) protected int
	TeammateTeamIndex,
	EnemyTeamIndex;

var(Tick) protected array<RePRI>
	Teammates, // An array of Teammates or the red team if bOnlySpectator.
	Enemies; // An array of enemies or the blue team if bOnlySpectator.

const MaxColumns = 8; // The columns that are drawn from right-to-left for each slot.


var(Opacity) GlobalConfig byte
	TeamBarOpacity,
	TeamScoreBarOpacity,
	TeamSummaryBarOpacity,
	NameOpacity,
	OwnNameOpacity,
	MatchStatusOpacity;


var(Colors) GlobalConfig Color
	CountryCodeColor,
	CountryFlagColor,
	TeamNeutralColor,
	LegendColor,
	RowColor,
	AdministratorColor,
	PortraitColor;


var(Tick) protected Color
	TopDataColor[MaxColumns],
	BottomDataColor[MaxColumns],
	TopLegendColor[MaxColumns],
	BottomLegendColor[MaxColumns];


var(ConstantStrings) protected localized const string
	TopLegendString[MaxColumns], // Ping / Kills / Score
	BottomLegendString[MaxColumns], // P/L / Deaths / PPH
	SlotAvailableBotString;

var(ConstantStrings) protected const string EmptyTopString[MaxColumns]; // "---"

var(Tick) protected string
	TopDataString[MaxColumns],
	BottomDataString[MaxColumns];

var protected float
	SlotColumn_Width[MaxColumns],
	Slot_OffsetX_Right[MaxColumns];

struct SCache
{
	var Font Font[2];

	var int Count_Rows;

	var float
		Width_Character[2],
		Height_Font[2],

		Width_Screen,
		Height_Screen,

		Width,

		Width_Slot,
		Height_Slot,
		X_Slot[IndexTeam_Max],

		Width_Status,
		OffsetX_Status,
		OffsetY_Status,
		Size_Status,

		Width_Portrait,
		OffsetX_Portrait,
		OffsetY_Portrait,
		Size_Portrait,

		Width_Flag,
		OffsetX_Flag,
		OffsetY_Flag,
		Size_Flag,

		OffsetX_CountrySymbolFlag,
		OffsetX_CountrySymbolText,

		OffsetX_PlayerName,
		Width_PlayerName,

		OffsetY_Middle,
		OffsetY_Top,
		OffsetY_Bottom;
};

var protected SCache Cached;

var(Tick) protected float Metrics_Total_Score[IndexTeam_Max];

var(Tick) protected int
	TeamPlayerCount[IndexTeam_Max],
	Metrics_Total_Kills[IndexTeam_Max],
	Metrics_Average_Ping[IndexTeam_Max],
	Metrics_Total_Ping[IndexTeam_Max];

var(ConstantTextures) protected const Texture
	LegendTexture,
	TeamScoreTexture,
	RowTexture,
	OwnRowTexture,
	SummaryTexture;

var(Colors) protected Color
	TeammateColor,
	EnemyColor,
	BalanceTeamsColor,
	EmptySlotColor;

var(Constants) protected const float
	ScaledWidth_Scoreboard,
	ScaledWidth_Slot,
	ScaledPositionX_Slot[IndexTeam_Max],
	ScaledPositionX_Scoreboard,
	ScaledPositionY_Scoreboard;

// #Metrics

protected function UpdateReset ()
{
	Players.Length = 0;

	Spectators.Length = 0;

	Teammates.Length = 0;
	Enemies.Length = 0;

	TeamPlayerCount[IndexTeam_A] = 0;
	TeamPlayerCount[IndexTeam_B] = 0;

	Metrics_Total_Kills[IndexTeam_A] = 0;
	Metrics_Total_Kills[IndexTeam_B] = 0;

	Metrics_Total_Ping[IndexTeam_A] = 0;
	Metrics_Total_Ping[IndexTeam_B] = 0;

	Metrics_Total_Score[IndexTeam_A] = 0.0f;
	Metrics_Total_Score[IndexTeam_B] = 0.0f;
}

// Preconditions:
// - (OtherRePRI.Team != None)
// - (OtherRePRI.TeamData != None)

protected function UpdatePlayer (RePRI OtherRePRI)
{
	local int Ping;

	Super.UpdatePlayer (OtherRePRI);

	if (OtherRePRI.Team.TeamIndex == TeammateTeamIndex)
	{
		Teammates[Teammates.Length] = OtherRePRI;

		if (!OtherRePRI.bBot)
		{
			Ping = Min (OtherRePRI.Ping * 4, 999);

			if (Ping != 999)
			{
				Metrics_Total_Ping[TeammateTeamIndex] += Ping;
			}

			++TeamPlayerCount[TeammateTeamIndex];
		}

		Metrics_Total_Kills[TeammateTeamIndex] += OtherRePRI.ReplicatedKills;
		Metrics_Total_Score[TeammateTeamIndex] += int(OtherRePRI.Score);
	}
	else
	{
		Enemies[Enemies.Length] = OtherRePRI;

		if (!OtherRePRI.bBot)
		{
			Ping = Min (OtherRePRI.Ping * 4, 999);

			if (Ping != 999)
			{
				Metrics_Total_Ping[EnemyTeamIndex] += Ping;
			}

			++TeamPlayerCount[EnemyTeamIndex];
		}

		Metrics_Total_Kills[EnemyTeamIndex] += OtherRePRI.ReplicatedKills;
		Metrics_Total_Score[EnemyTeamIndex] += int(OtherRePRI.Score);
	}
}

// Protocol:
// - Return True if there was enough information to update metrics.
// Preconditions:
// - (GRI != None)

protected function bool UpdateScoreboard_Metrics ()
{
	local PlayerReplicationInfo OtherRePRI;
	local int Index;
	local int Count;

	Count = GRI.PRIArray.Length;

	// OwnTeamIndex

	if (PlayerOwnerPRI.bOnlySpectator)
	{
		OwnTeamIndex = IndexTeam_Unspecified;

		TeammateTeamIndex = IndexTeam_A;
		EnemyTeamIndex = IndexTeam_B;

		TeammateColor = Class'PlayerSettings'.Default.StandardTeamColor[TeammateTeamIndex];
		EnemyColor = Class'PlayerSettings'.Default.StandardTeamColor[EnemyTeamIndex];
	}
	else if (PlayerOwnerPRI.Team != None)
	{
		OwnTeamIndex = PlayerOwnerPRI.Team.TeamIndex;

		if (OwnTeamIndex == IndexTeam_A)
		{
			TeammateTeamIndex = IndexTeam_A;
			EnemyTeamIndex = IndexTeam_B;
		}
		else if (OwnTeamIndex == IndexTeam_B)
		{
			TeammateTeamIndex = IndexTeam_B;
			EnemyTeamIndex = IndexTeam_A;
		}

		TeammateColor = Class'PlayerSettings'.Static.GetCustomTeammateColor (TeammateTeamIndex);
		EnemyColor = Class'PlayerSettings'.Static.GetCustomEnemyColor (EnemyTeamIndex);
	}
	else // Assume these defaults until replication gets client updated.
	{
		OwnTeamIndex = IndexTeam_Unspecified;

		TeammateTeamIndex = IndexTeam_A;
		EnemyTeamIndex = IndexTeam_B;

		TeammateColor = TeamNeutralColor;
		EnemyColor = TeamNeutralColor;
	}

	// Data

	UpdateReset ();

	// Players

	for (Index = 0; Index < GRI.PRIArray.Length; ++Index)
	{
		OtherRePRI = GRI.PRIArray[Index];

		if (OtherRePRI != None)
		{
			if (OtherRePRI.bOnlySpectator)
			{
				UpdateSpectator (OtherRePRI);
			}
			else if
			(
				OtherRePRI.IsA ('RePRI')
				&&
				(OtherRePRI.Team != None)
			)
			{
				UpdatePlayer (RePRI(OtherRePRI));
			}
		}
	}

	// <NM_Client>
	if (TeamPlayerCount[TeammateTeamIndex] > 0)
	{
		Metrics_Average_Ping[TeammateTeamIndex] = Metrics_Total_Ping[TeammateTeamIndex] / TeamPlayerCount[TeammateTeamIndex];
	}

	if (TeamPlayerCount[EnemyTeamIndex] > 0)
	{
		Metrics_Average_Ping[EnemyTeamIndex] = Metrics_Total_Ping[EnemyTeamIndex] / TeamPlayerCount[EnemyTeamIndex];
	}
	// </NM_Client>

	return True;
}

event UpdateScoreboard (Canvas Canvas)
{
	local float Y;
	local int IndexRow;
	local int Count_Rows;

	// Figure out how many rows are needed to show all players.

	Count_Rows
	=
	Clamp
	(
		Max
		(
			Teammates.Length,
			Enemies.Length
		),
		MinSlotsDrawn,
		MaxSlotsDrawn
	);

	if
	(
		(ReGRI(GRI) != None)
		&&
		!GRI.bMatchHasBegun
	)
	{
		Count_Rows = Max ((ReGRI(GRI).MaxPlayers / 2), Count_Rows);
	}

	// Recalculate when number of players change or screen resolution changes.

	if
	(
		(Count_Rows != Cached.Count_Rows)
		||
		(Canvas.ClipX != Cached.Width_Screen)
		||
		(Canvas.ClipY != Cached.Height_Screen)
	)
	{
		UpdateScoreboard_Precalc (Canvas);
	}

	// Start Drawing

	Canvas.Style = ERenderStyle.STY_Alpha;

	if (Level.NetMode != NM_Standalone)
	{
		Y = DrawHeader (Canvas, Y);
    }
    else
    {
		Y = Canvas.ClipY * 0.04f;
    }

	Y = DrawMatchStatus (Canvas, Y + 4.0f);
	Y = DrawTeamScores (Canvas, Y + 4.0f);
	Y = DrawTeamLegends (Canvas, Y);
	Y = DrawTeamBars (Canvas, Y);

	DrawMatchInfo (Canvas);

	// <Scoreboard_Draw_Slots (Canvas)>
	// - 1. Row
	// - 2. (Teammates || Red) | (Enemies || Blue)

	for (IndexRow = 0; IndexRow < Count_Rows; ++IndexRow)
	{
		if (IndexRow < Teammates.Length)
		{
			DrawPlayerSlot
			(
				Canvas,
				Teammates[IndexRow],
				Cached.X_Slot[0],
				Y,
				TeammateTeamIndex,
				True
			);
		}
		else
		{
			DrawEmptySlot (Canvas, Cached.X_Slot[0], Y);
		}

		if (IndexRow < Enemies.Length)
		{
			DrawPlayerSlot
			(
				Canvas,
				Enemies[IndexRow],
				Cached.X_Slot[1],
				Y,
				EnemyTeamIndex,
				(PlayerOwnerPRI.bOnlySpectator || PlayerOwnerPRI.bAdmin)
			);
		}
		else
		{
			DrawEmptySlot (Canvas, Cached.X_Slot[1], Y);
		}


		Y += Cached.Height_Slot;
	}
	// </Scoreboard_Draw_Slots (Canvas)>


	if
	(
		(GRI.Teams[TeammateTeamIndex] != None)
		&&
		(GRI.Teams[EnemyTeamIndex] != None)
	)
	{
		Y = DrawTeamSummaries (Canvas, Y);


		Y = DrawMatchRules (Canvas, Y + 8.0f);


		if (Class'PlayerSettings'.Default.bShowScoreboardVersus)
		{
			Y = DrawVersusInfo (Canvas, Y + 4.0f);
		}
	}


	if
	(
		(Class'PlayerSettings'.Default.bShowScoreboardSpectators)
		&&
		(Spectators.Length > 0)
	)
	{
		Y = DrawSpectators (Canvas, Y + 4.0f);
	}
}

// - Any change in screen resolution or player count will trigger
// a recalculation.

protected function UpdateScoreboard_Precalc (Canvas Canvas)
{
	local float SizeScaled_Large, SizeScaled_Small, Width_Upper, Width_Highest, StringHeight, OffsetX;
	local int SizeFont_Base;

	Cached.Count_Rows
	=
	Clamp
	(
		Max
		(
			Teammates.Length,
			Enemies.Length
		),
		MinSlotsDrawn,
		MaxSlotsDrawn
	);

	if
	(
		(ReGRI(GRI) != None)
		&&
		!GRI.bMatchHasBegun
	)
	{
		Cached.Count_Rows = Max ((ReGRI(GRI).MaxPlayers / 2), Cached.Count_Rows);
	}

	// Base Font Size

	if (Cached.Count_Rows > 13)
	{
		SizeFont_Base = -5;
	}
	else if (Cached.Count_Rows > 11)
	{
		SizeFont_Base = -4;
	}
	else if (Cached.Count_Rows > 9)
	{
		SizeFont_Base = -3;
	}
	else
	{
		SizeFont_Base = -2;
	}

	//

	Cached.Font[SizeFont_Small] = PlayerOwnerHUD.GetFontSizeIndex (Canvas, SizeFont_Base - 1);
	Canvas.Font = Cached.Font[SizeFont_Small];
	Canvas.TextSize ("1", Cached.Width_Character[SizeFont_Small], Cached.Height_Font[SizeFont_Small]);

	Cached.Font[SizeFont_Large] = PlayerOwnerHUD.GetFontSizeIndex (Canvas, SizeFont_Base);
	Canvas.Font = Cached.Font[SizeFont_Large];
	Canvas.TextSize ("1", Cached.Width_Character[SizeFont_Large], Cached.Height_Font[SizeFont_Large]);

	//

	Cached.Width_Screen = Canvas.ClipX;
	Cached.Height_Screen = Canvas.ClipY;

	//

	Cached.Width = Canvas.ClipX * ScaledWidth_Scoreboard;
	Cached.Width_Slot = Canvas.ClipX * ScaledWidth_Slot;

	//

	SizeScaled_Large = 1.3f;
	SizeScaled_Small = 1.3f;

	Cached.Height_Slot = FMax (Cached.Height_Font[SizeFont_Large] * SizeScaled_Large, Cached.Height_Font[SizeFont_Small] * 2.0f * SizeScaled_Small);

	//

	Cached.X_Slot[0] = Cached.Width_Screen * ScaledPositionX_Slot[0];
	Cached.X_Slot[1] = Cached.Width_Screen * ScaledPositionX_Slot[1];

	Cached.OffsetY_Middle = (Cached.Height_Slot * 0.5f) - (Cached.Height_Font[SizeFont_Large] * 0.5f);
	Cached.OffsetY_Top = (Cached.Height_Slot * 0.25f) - (Cached.Height_Font[1] * 0.5f);
	Cached.OffsetY_Bottom = (Cached.Height_Slot * 0.75f) - (Cached.Height_Font[1] * 0.5f);

	Cached.Width_Status = Cached.Height_Slot;
	Cached.Size_Status = 48.0f * (Cached.Height_Slot / 64.0f);
	Cached.OffsetX_Status = (Cached.Width_Slot * 0.5f) - (Cached.Size_Status * 0.5f);
	Cached.OffsetY_Status = (Cached.Width_Status * 0.5f) - (Cached.Size_Status * 0.5f); // Cached.OffsetX_Status;

	Cached.Width_Flag = Cached.Width_Character[0] + Cached.Height_Slot;
	Cached.Size_Flag = 60.0f * (Cached.Height_Slot / 64.0f);
	Cached.OffsetX_Flag = (Cached.Width_Flag * 0.5f) - (Cached.Size_Flag * 0.5f);
	Cached.OffsetY_Flag = (Cached.Height_Slot * 0.5f) - (Cached.Size_Flag * 0.5f);

	Cached.OffsetX_CountrySymbolText = ((Cached.Width_Status + Cached.Width_Character[0]) * 0.5f) - Cached.Width_Character[SizeFont_Large];

	Cached.Width_Portrait = Cached.Height_Slot;
	Cached.Size_Portrait = Cached.Height_Slot;
	Cached.OffsetX_Portrait = 0.0f;
	Cached.OffsetY_Portrait = (Cached.Height_Slot * 0.5f) - (Cached.Size_Portrait * 0.5f);

	Cached.Width_PlayerName = Cached.Width_Character[0] * 20;
	Cached.OffsetX_PlayerName = Cached.Width_Status + Cached.Width_Character[0];

	// 1 Ping

	OffsetX = Cached.Width_Slot;

	if (Level.NetMode == NM_Client)
	{
		Canvas.TextSize (TopLegendString[0], Width_Upper, StringHeight);
		Canvas.TextSize (BottomLegendString[0], SlotColumn_Width[0], StringHeight);

		if (Width_Upper > SlotColumn_Width[0])
		{
			SlotColumn_Width[0] = Width_Upper;
		}

		Canvas.TextSize (EmptyTopString[0], Width_Highest, StringHeight);

		if (Width_Highest > SlotColumn_Width[0])
		{
			SlotColumn_Width[0] = Width_Highest;
		}

		SlotColumn_Width[0] += Cached.Width_Character[SizeFont_Large];
		OffsetX -= SlotColumn_Width[0];
		Slot_OffsetX_Right[0] = OffsetX + (SlotColumn_Width[0] * 0.5f);
	}

	// 2 Kills

	Canvas.TextSize (TopLegendString[1], Width_Upper, StringHeight);
	Canvas.TextSize (BottomLegendString[1], SlotColumn_Width[1], StringHeight);

	if (Width_Upper > SlotColumn_Width[1])
	{
		SlotColumn_Width[1] = Width_Upper;
	}

	Canvas.TextSize (EmptyTopString[1], Width_Highest, StringHeight);

	if (Width_Highest > SlotColumn_Width[1])
	{
		SlotColumn_Width[1] = Width_Highest;
	}

	SlotColumn_Width[1] += Cached.Width_Character[SizeFont_Large];
	OffsetX -= SlotColumn_Width[1];
	Slot_OffsetX_Right[1] = OffsetX + (SlotColumn_Width[1] * 0.5f);

	// 3 Score

	Canvas.TextSize (TopLegendString[2], Width_Upper, StringHeight);
	Canvas.TextSize (BottomLegendString[2], SlotColumn_Width[2], StringHeight);

	if (Width_Upper > SlotColumn_Width[2])
	{
		SlotColumn_Width[2] = Width_Upper;
	}

	Canvas.TextSize (EmptyTopString[2], Width_Highest, StringHeight);

	if (Width_Highest > SlotColumn_Width[2])
	{
		SlotColumn_Width[2] = Width_Highest;
	}

	SlotColumn_Width[2] += Cached.Width_Character[SizeFont_Large];
	OffsetX -= SlotColumn_Width[2];
	Slot_OffsetX_Right[2] = OffsetX + (SlotColumn_Width[2] * 0.5f);

	// 4

	Canvas.TextSize (TopLegendString[3], Width_Upper, StringHeight);
	Canvas.TextSize (BottomLegendString[3], SlotColumn_Width[3], StringHeight);

	if (Width_Upper > SlotColumn_Width[3])
	{
		SlotColumn_Width[3] = Width_Upper;
	}

	Canvas.TextSize (EmptyTopString[3], Width_Highest, StringHeight);

	if (Width_Highest > SlotColumn_Width[3])
	{
		SlotColumn_Width[3] = Width_Highest;
	}

	SlotColumn_Width[3] += Cached.Width_Character[SizeFont_Large];
	OffsetX -= SlotColumn_Width[3];
	Slot_OffsetX_Right[3] = OffsetX + (SlotColumn_Width[3] * 0.5f);
}

// #Rendering

// - Draws Player, Score, Kills, Deaths, (...).

protected function float DrawTeamLegends
(
	Canvas Canvas,
	float Y
)
{
	local float StringWidth, StringHeight;

	// <DrawTeams>
	Canvas.DrawColor = LegendColor;

	Canvas.SetPos (Cached.X_Slot[0], Y);

	Canvas.DrawTile
	(
		LegendTexture,
		Cached.Width_Slot,
		Cached.Height_Slot,
		0.0f,
		0.0f,
		LegendTexture.USize - 1.0f,
		LegendTexture.VSize - 1.0f
	);


	Canvas.SetPos (Cached.X_Slot[1], Y);

	Canvas.DrawTile
	(
		LegendTexture,
		Cached.Width_Slot,
		Cached.Height_Slot,
		0.0f,
		0.0f,
		LegendTexture.USize - 1.0f,
		LegendTexture.VSize - 1.0f
	);
	// </DrawTeams>


	// <DrawPlayers>
	Canvas.Font = Cached.Font[SizeFont_Large];

	Canvas.DrawColor = WhiteColor;

	Canvas.SetPos
	(
		Cached.X_Slot[0] + Cached.OffsetX_PlayerName,
		Y + Cached.OffsetY_Middle
	);

	Canvas.DrawText (PlayerString, False);


	Canvas.SetPos
	(
		Cached.X_Slot[1] + Cached.OffsetX_PlayerName,
		Y + Cached.OffsetY_Middle
	);

	Canvas.DrawText (PlayerString, False);
	// </DrawPlayers>


	// Upper

	if (GRI.bMatchhasBegun)
	{
		// <DrawPing>
		if (Level.NetMode == NM_Client)
		{
			Canvas.DrawColor = TopLegendColor[0];

			Canvas.TextSize (TopLegendString[0], StringWidth, StringHeight);

			StringWidth *= 0.5f;

			Canvas.SetPos
			(
				Cached.X_Slot[0] + Slot_OffsetX_Right[0] - StringWidth,
				Y + Cached.OffsetY_Top
			);

			Canvas.DrawText (TopLegendString[0], False);


			Canvas.SetPos
			(
				Cached.X_Slot[1] + Slot_OffsetX_Right[0] - StringWidth,
				Y + Cached.OffsetY_Top
			);

			Canvas.DrawText (TopLegendString[0], False);
		}
		// </DrawPing>


		// <DrawKills>
		Canvas.DrawColor = TopLegendColor[1];

		Canvas.TextSize (TopLegendString[1], StringWidth, StringHeight);

		StringWidth *= 0.5f;

		Canvas.SetPos
		(
			Cached.X_Slot[0] + Slot_OffsetX_Right[1] - StringWidth,
			Y + Cached.OffsetY_Top
		);

		Canvas.DrawText (TopLegendString[1], False);


		Canvas.SetPos
		(
			Cached.X_Slot[1] + Slot_OffsetX_Right[1] - StringWidth,
			Y + Cached.OffsetY_Top
		);

		Canvas.DrawText (TopLegendString[1], False);
		// </DrawKills>


		// <DrawScore>
		if (TopLegendString[2] != "")
		{
			Canvas.DrawColor = TopLegendColor[2];

			Canvas.TextSize (TopLegendString[2], StringWidth, StringHeight);

			StringWidth *= 0.5f;

			Canvas.SetPos
			(
				Cached.X_Slot[0] + Slot_OffsetX_Right[2] - StringWidth,
				Y + Cached.OffsetY_Top
			);

			Canvas.DrawText (TopLegendString[2], False);


			Canvas.SetPos
			(
				Cached.X_Slot[1] + Slot_OffsetX_Right[2] - StringWidth,
				Y + Cached.OffsetY_Top
			);

			Canvas.DrawText (TopLegendString[2], False);
		}
		// </DrawScore>


		// <4>
		if (TopLegendString[3] != "")
		{
			Canvas.DrawColor = TopLegendColor[3];

			Canvas.TextSize (TopLegendString[3], StringWidth, StringHeight);

			StringWidth *= 0.5f;

			Canvas.SetPos
			(
				Cached.X_Slot[0] + Slot_OffsetX_Right[3] - StringWidth,
				Y + Cached.OffsetY_Top
			);

			Canvas.DrawText (TopLegendString[3], False);


			Canvas.SetPos
			(
				Cached.X_Slot[1] + Slot_OffsetX_Right[3] - StringWidth,
				Y + Cached.OffsetY_Top
			);

			Canvas.DrawText (TopLegendString[3], False);
		}
		// </4>


		Canvas.Font = Cached.Font[SizeFont_Small];


		// <DrawPacketLoss>
		if (Level.NetMode == NM_Client)
		{
			Canvas.DrawColor = BottomLegendColor[0];

			Canvas.TextSize (BottomLegendString[0], StringWidth, StringHeight);

			StringWidth *= 0.5f;

			Canvas.SetPos
			(
				Cached.X_Slot[0] + Slot_OffsetX_Right[0] - StringWidth,
				Y + Cached.OffsetY_Bottom
			);

			Canvas.DrawText (BottomLegendString[0], False);


			Canvas.SetPos
			(
				Cached.X_Slot[1] + Slot_OffsetX_Right[0] - StringWidth,
				Y + Cached.OffsetY_Bottom
			);

			Canvas.DrawText (BottomLegendString[0], False);
		}
		// </DrawPacketLoss>


		// <DrawDeaths>
		Canvas.DrawColor = BottomLegendColor[1];

		Canvas.TextSize (BottomLegendString[1], StringWidth, StringHeight);

		StringWidth *= 0.5f;

		Canvas.SetPos
		(
			Cached.X_Slot[0] + Slot_OffsetX_Right[1] - StringWidth,
			Y + Cached.OffsetY_Bottom
		);

		Canvas.DrawText (BottomLegendString[1], False);


		Canvas.SetPos
		(
			Cached.X_Slot[1] + Slot_OffsetX_Right[1] - StringWidth,
			Y + Cached.OffsetY_Bottom
		);

		Canvas.DrawText (BottomLegendString[1], False);
		// </DrawDeaths>


		// <DrawPointsPerRound>
		Canvas.DrawColor = BottomLegendColor[2];

		Canvas.TextSize (BottomLegendString[2], StringWidth, StringHeight);

		StringWidth *= 0.5f;

		Canvas.SetPos
		(
			Cached.X_Slot[0] + Slot_OffsetX_Right[2] - StringWidth,
			Y + Cached.OffsetY_Bottom
		);

		Canvas.DrawText (BottomLegendString[2], False);


		Canvas.SetPos
		(
			Cached.X_Slot[1] + Slot_OffsetX_Right[2] - StringWidth,
			Y + Cached.OffsetY_Bottom
		);

		Canvas.DrawText (BottomLegendString[2], False);
		// </DrawPointsPerRound>


		// <3>
		if (BottomLegendString[3] != "")
		{
			Canvas.DrawColor = BottomLegendColor[3];

			Canvas.TextSize (BottomLegendString[3], StringWidth, StringHeight);

			StringWidth *= 0.5f;

			Canvas.SetPos
			(
				Cached.X_Slot[0] + Slot_OffsetX_Right[3] - StringWidth,
				Y + Cached.OffsetY_Bottom
			);

			Canvas.DrawText (BottomLegendString[3], False);


			Canvas.SetPos
			(
				Cached.X_Slot[1] + Slot_OffsetX_Right[3] - StringWidth,
				Y + Cached.OffsetY_Bottom
			);

			Canvas.DrawText (BottomLegendString[3], False);
		}
		// </3>
	}
	else
	{
		if (Level.NetMode == NM_Client)
		{
			// <DrawPing>
			Canvas.DrawColor = TopLegendColor[0];

			Canvas.TextSize (TopLegendString[0], StringWidth, StringHeight);

			StringWidth *= 0.5f;

			Canvas.SetPos
			(
				Cached.X_Slot[0] + Slot_OffsetX_Right[0] - StringWidth,
				Y + Cached.OffsetY_Top
			);

			Canvas.DrawText (TopLegendString[0], False);


			Canvas.SetPos
			(
				Cached.X_Slot[1] + Slot_OffsetX_Right[0] - StringWidth,
				Y + Cached.OffsetY_Top
			);

			Canvas.DrawText (TopLegendString[0], False);
			// </DrawPing>


			// <DrawPacketLoss>
			Canvas.DrawColor = BottomLegendColor[0];

			Canvas.Font = Cached.Font[SizeFont_Small];

			Canvas.TextSize (BottomLegendString[0], StringWidth, StringHeight);

			StringWidth *= 0.5f;

			Canvas.SetPos
			(
				Cached.X_Slot[0] + Slot_OffsetX_Right[0] - StringWidth,
				Y + Cached.OffsetY_Bottom
			);

			Canvas.DrawText (BottomLegendString[0], False);

			Canvas.SetPos
			(
				Cached.X_Slot[1] + Slot_OffsetX_Right[0] - StringWidth,
				Y + Cached.OffsetY_Bottom
			);

			Canvas.DrawText (BottomLegendString[0], False);
			// </DrawPacketLoss>
		}
	}

	// Finalize

	return (Y + Cached.Height_Slot);
}

// These bars are drawn below the legend and before the first player.

protected function float DrawTeamScores
(
	Canvas Canvas,
	float Y
)
{
	local string CachedString;
	local float WidthString, HeightString, Width_Bar, Height_Bar, OffsetX_Bar, Y_Text;

	// Common

	Canvas.Font = HUDClass.Static.GetMediumFontfor (Canvas);

	Canvas.TextSize ("-*****-", WidthString, HeightString);

	Width_Bar = WidthString;
	Height_Bar = HeightString + 4.0f;
	OffsetX_Bar = (Cached.Width_Slot * 0.5f) - (Width_Bar * 0.5f);
	Y_Text = Y + (Height_Bar * 0.5f) - (HeightString * 0.5f);

	// Your Team

	if (GRI.Teams[TeammateTeamIndex] != None)
	{
		CachedString = string(int(GRI.Teams[TeammateTeamIndex].Score));
	}
	else
	{
		CachedString = "0";
	}

	Canvas.TextSize (CachedString, WidthString, HeightString);

	// Your Team Bar

	Canvas.DrawColor = Class'PlayerSettings'.Static.GetRelativeTeamColor (OwnTeamIndex, TeammateTeamIndex);
	Canvas.DrawColor.A = TeamScoreBarOpacity;

	Canvas.SetPos (Cached.X_Slot[0] + OffsetX_Bar, Y);

	Canvas.DrawTile
	(
		TeamScoreTexture,
		Width_Bar,
		Height_Bar,
		0.0f,
		0.0f,
		TeamScoreTexture.USize - 1.0f,
		TeamScoreTexture.VSize - 1.0f
	);

	// Your Team Score

	Canvas.DrawColor = WhiteColor;

	Canvas.SetPos
	(
		Cached.X_Slot[0] + (Cached.Width_Slot * 0.5f) - (WidthString * 0.5f),
		Y_Text
	);

	Canvas.DrawText (CachedString, False);

	// Enemy Team

	if (GRI.Teams[EnemyTeamIndex] != None)
	{
		CachedString = string(int(GRI.Teams[EnemyTeamIndex].Score));
	}
	else
	{
		CachedString = "0";
	}

	Canvas.TextSize (CachedString, WidthString, HeightString);


	Canvas.DrawColor = WhiteColor;

	Canvas.SetPos
	(
		Cached.X_Slot[1] + (Cached.Width_Slot * 0.5f) - (WidthString * 0.5f),
		Y_Text
	);

	Canvas.DrawText (CachedString, False);

	// Enemy Team Bar

	Canvas.DrawColor = Class'PlayerSettings'.Static.GetRelativeTeamColor (OwnTeamIndex, EnemyTeamIndex);
	Canvas.DrawColor.A = TeamScoreBarOpacity;

	Canvas.SetPos (Cached.X_Slot[1] + OffsetX_Bar, Y);

	Canvas.DrawTile
	(
		TeamScoreTexture,
		Width_Bar,
		Height_Bar,
		0.0f,
		0.0f,
		TeamScoreTexture.USize - 1.0f,
		TeamScoreTexture.VSize - 1.0f
	);

	// Enemy Team Score

	Canvas.DrawColor = WhiteColor;

	Canvas.SetPos
	(
		Cached.X_Slot[1] + (Cached.Width_Slot * 0.5f) - (WidthString * 0.5f),
		Y_Text
	);

	Canvas.DrawText (CachedString, False);

	//

	return (Y + Height_Bar);
}

// Draws a team colored bar

protected function float DrawTeamBars
(
	Canvas Canvas,
	float Y
)
{
	// Your Team Bar

	Canvas.DrawColor = Class'PlayerSettings'.Static.GetRelativeTeamColor (OwnTeamIndex, TeammateTeamIndex);
	Canvas.DrawColor.A = TeamBarOpacity;

	Canvas.SetPos (Cached.X_Slot[0], Y);

	Canvas.DrawTilePartialStretched
	(
		TeamScoreTexture,
		Cached.Width_Slot,
		Cached.Height_Font[SizeFont_Small]
	);

	// Enemy Team Bar

	Canvas.DrawColor = Class'PlayerSettings'.Static.GetRelativeTeamColor (OwnTeamIndex, EnemyTeamIndex);
	Canvas.DrawColor.A = TeamBarOpacity;

	Canvas.SetPos (Cached.X_Slot[1], Y);

	Canvas.DrawTilePartialStretched
	(
		TeamScoreTexture,
		Cached.Width_Slot,
		Cached.Height_Font[SizeFont_Small]
	);

	//

	return (Y + Cached.Height_Font[SizeFont_Small]);
}

// #Slot

// - Column zero is the first column drawn starting from the right side of the slot.
// - Reminder: Check this function in derived classes.

protected function UpdatePlayerSlotData (PlayerReplicationInfo OtherPRI)
{
	// Right-Most Column

	if (Level.NetMode == NM_Client)
	{
		if (!OtherPRI.bBot)
		{
			TopDataColor[0] = TopLegendColor[0];

			if (!bMatchOver)
			{
				TopDataString[0] = string(Min (OtherPRI.Ping * 4, 999));
			}
			else
			{
				TopDataString[0] = "---";
			}


			if (OtherPRI.PacketLoss > 0)
			{
				BottomDataColor[0] = ExperiencingPacketlossColor;
			}
			else if (RePRI(OtherPRI).bHadPacketLoss)
			{
				BottomDataColor[0] = HadPacketlossColor;
			}
			else
			{
				BottomDataColor[0] = ConnectionColor;
			}

			BottomDataString[0] = string(OtherPRI.PacketLoss);
		}
		else
		{
			TopDataColor[0] = TopLegendColor[0];
			TopDataString[0] = BotConnectionString;

			BottomDataColor[0] = BottomLegendColor[0];
			BottomDataString[0] = BotConnectionString;
		}

		// KILLS / DEATHS

		TopDataColor[1] = TopLegendColor[1];
		TopDataString[1] = string(RePRI(OtherPRI).ReplicatedKills);

		BottomDataColor[1] = BottomLegendColor[1];
		BottomDataString[1] = string(int(OtherPRI.Deaths));

		// SCORE / PPH

		TopDataColor[2] = TopLegendColor[2];
		TopDataString[2] = string(int(OtherPRI.Score));
		BottomDataColor[2] = WhiteColor;

		if
		(
			!UnrealPlayer(Owner).bDisplayLoser
			&&
			!UnrealPlayer(Owner).bDisplayWinner
			&&
			(GRI.ElapsedTime > 0)
		)
		{
			BottomDataString[2] = string(Clamp (3600 * OtherPRI.Score / FMax (1.0f, GRI.ElapsedTime - OtherPRI.StartTime), -999.0f, 9999.0f));
		}
		else
		{
			BottomDataString[2] = "---";
		}
	}
	else
	{
		// KILLS

		TopDataColor[1] = TopLegendColor[1];
		TopDataString[1] = string(RePRI(OtherPRI).ReplicatedKills);

		// DEATHS

		BottomDataColor[1] = BottomLegendColor[1];
		BottomDataString[1] = string(int(OtherPRI.Deaths));

		// SCORE

		TopDataColor[2] = TopLegendColor[2];
		TopDataString[2] = string(int(OtherPRI.Score));

		// PPH

		BottomDataColor[2] = BottomLegendColor[2];

		if
		(
			!UnrealPlayer(Owner).bDisplayLoser
			&&
			!UnrealPlayer(Owner).bDisplayWinner
			&&
			(GRI.ElapsedTime > 0)
		)
		{
			BottomDataString[2] = string(Clamp (3600 * OtherPRI.Score / FMax (1.0f, GRI.ElapsedTime - OtherPRI.StartTime), -999.0f, 9999.0f));
		}
		else
		{
			BottomDataString[2] = "---";
		}
	}
}

// - Draws player country code, flag or portrait, player name, location or status,
// and calls a function to draw player data (score, kills, deaths, ...).

protected function DrawPlayerSlot
(
	Canvas Canvas,
	PlayerReplicationInfo OtherPRI,
	float X,
	float Y,
	int IndexTeam,
	bool bDrawLocation
)
{
	local Material PortraitMaterial;
	local string CachedString;
	local bool bDrawFlag;

	// <DrawPortrait>
	bDrawFlag = !Class'PlayerSettings'.Default.bScoreboardShowPlayerPortrait;

	if
	(
		!bDrawFlag // If portrait wanted, or blank country code, ...
		||
		(RePRI(OtherPRI).CountryCode == "")
	)
	{
		Canvas.DrawColor = PortraitColor;

		Canvas.SetPos
		(
			X + Cached.OffsetX_Portrait,
			Y + Cached.OffsetY_Portrait
		);

		PortraitMaterial = OtherPRI.GetPortrait ();

		Canvas.DrawTile
		(
			PortraitMaterial,
			Cached.Size_Portrait,
			Cached.Size_Portrait,
			0.0f,
			0.0f,
			PortraitMaterial.MaterialUSize (),
			PortraitMaterial.MaterialUSize ()
		);
	}
	// </DrawPortrait>

 	// <DrawRowBackground>
	if (!OtherPRI.bAdmin) // Not admin
	{
		if (PlayerOwnerPRI != OtherPRI) // Not local player
		{
			if (OtherPRI != SpectatedPRI) // Not spectated player
			{
				Canvas.DrawColor = RowColor;
            }
            else
            {
				Canvas.DrawColor = SpectatedRowColor;
            }

			Canvas.SetPos (X, Y);

			Canvas.DrawTile
			(
				RowTexture,
				Cached.Width_Slot,
				Cached.Height_Slot,
				0.0f,
				0.0f,
				RowTexture.USize - 1.0f,
				RowTexture.VSize - 1.0f
			);
		}
		else // Local player
		{
			Canvas.DrawColor = TeammateColor;
			Canvas.DrawColor.A = OwnNameOpacity;

			Canvas.SetPos (X, Y);

			Canvas.DrawTilePartialStretched
			(
				OwnRowTexture,
				Cached.Width_Slot,
				Cached.Height_Slot
			);
		}
	}
	else
	{
		Canvas.DrawColor = AdministratorColor;

		Canvas.SetPos (X, Y);

		Canvas.DrawTilePartialStretched
		(
			OwnRowTexture,
			Cached.Width_Slot,
			Cached.Height_Slot
		);
	}
 	// </DrawRowBackground>

	// Portrait

	if
	(
		bDrawFlag
		&&
		(RePRI(OtherPRI).CountryCode != "")
	)
	{
		if (RePRI(OtherPRI).TextureFlag == None)
		{
			Class'CountryCodes'.Static.GetTextureFlag (RePRI(OtherPRI).CountryCode, RePRI(OtherPRI).TextureFlag, RePRI(OtherPRI).OffsetTextureFlag);
		}


		if (RePRI(OtherPRI).TextureFlag != None)
		{
			Canvas.DrawColor = CountryFlagColor;

			Canvas.SetPos
			(
				X + Cached.OffsetX_Flag,
				Y + Cached.OffsetY_Flag
			);

			Canvas.DrawTile
			(
				RePRI(OtherPRI).TextureFlag,
				Cached.Size_Flag,
				Cached.Size_Flag,
				RePRI(OtherPRI).OffsetTextureFlag.X1,
				RePRI(OtherPRI).OffsetTextureFlag.Y1,
				RePRI(OtherPRI).OffsetTextureFlag.X2,
				RePRI(OtherPRI).OffsetTextureFlag.Y2
			);
		}
		else
		{
			Canvas.DrawColor = CountryCodeColor;

			CachedString = RePRI(OtherPRI).GetCountryCodeString ();

			Canvas.SetPos
			(
				X + Cached.OffsetX_CountrySymbolText,
				Y + Cached.OffsetY_Middle
			);

			Canvas.DrawText (CachedString, False);
		}
	}


	// <DrawPlayerName>
	Canvas.Font = Cached.Font[SizeFont_Large];

	if (PlayerOwnerPRI != OtherPRI) // Not local player
	{
		if (SpectatedPRI != OtherPRI) // Not spectated player
		{
			if (!OtherPRI.bOutofLives)
			{
				Canvas.DrawColor = WhiteColor;
            }
            else
            {
				Canvas.DrawColor = DimmedWhiteColor;
            }
	    }
		else
	    {
			Canvas.DrawColor = SpectatedColor;
	    }
    }
    else
    {
		Canvas.DrawColor = Class'PlayerSettings'.Static.GetCurrentCustomPlayerColor ();
	}

	Canvas.SetPos
	(
		X + Cached.OffsetX_PlayerName,
		Y + Cached.OffsetY_Top
	);

	Canvas.DrawText (OtherPRI.PlayerName, False);
	// </DrawPlayerName>


	Canvas.Font = Cached.Font[SizeFont_Small];

	SetPlayerStatusColor (Canvas, OtherPRI);

	Canvas.SetPos
	(
		X + Cached.OffsetX_PlayerName + 4.0f,
		Y + Cached.OffsetY_Bottom
	);

	Canvas.DrawText (GetPlayerStatusString (OtherPRI, bDrawLocation), False);
	// </DrawLocationName>


	UpdatePlayerSlotData (OtherPRI);


	DrawPlayerSlotData (Canvas, OtherPRI, X, Y);
}


protected function string GetPlayerStatusString (PlayerReplicationInfo OtherPRI, bool bDrawLocation)
{
	if (GRI.bMatchhasBegun)
	{
		if (RePRI(OtherPRI).IsCurrentPlayerStatus ('Out'))
		{
			return OutString;
		}
		else if (RePRI(OtherPRI).IsCurrentPlayerStatus ('Dead'))
		{
			return DeadString;
		}
		else if (bDrawLocation)
		{
			return OtherPRI.GetLocationName ();
		}
		else
		{
			return Class'Engine.PlayerReplicationInfo'.Default.StringUnknown;
        }
	}
	else if (OtherPRI.bReadyToPlay)
	{
		return ReadyText;
    }
    else
    {
		return NotReadyText;
    }
}


protected function DrawPlayerSlotData (Canvas Canvas, PlayerReplicationInfo OtherPRI, float X, float Y)
{
	local float StringWidth, StringHeight;

	if (GRI.bMatchhasBegun)
	{
		if (Level.NetMode == NM_Client)
		{
			Canvas.Font = Cached.Font[SizeFont_Large];


			// <DrawPing>
			Canvas.DrawColor = TopDataColor[0];

			Canvas.TextSize (TopDataString[0], StringWidth, StringHeight);

			Canvas.SetPos
			(
				X + Slot_OffsetX_Right[0] - (StringWidth * 0.5f),
				Y + Cached.OffsetY_Top
			);

			Canvas.DrawText (TopDataString[0], False);
			// </DrawPing>


			// <Kills>
			Canvas.DrawColor = TopDataColor[1];

			Canvas.TextSize (TopDataString[1], StringWidth, StringHeight);

			Canvas.SetPos
			(
				X + Slot_OffsetX_Right[1] - (StringWidth * 0.5f),
				Y + Cached.OffsetY_Top
			);

			Canvas.DrawText (TopDataString[1], False);
			// </Kills>


			// <Score>
			Canvas.DrawColor = TopDataColor[2];

			Canvas.TextSize (TopDataString[2], StringWidth, StringHeight);

			Canvas.SetPos
			(
				X + Slot_OffsetX_Right[2] - (StringWidth * 0.5f),
				Y + Cached.OffsetY_Top
			);

			Canvas.DrawText (TopDataString[2], False);
			// </Score>


			Canvas.Font = Cached.Font[SizeFont_Small];


			// <PacketLoss>
			Canvas.DrawColor = BottomDataColor[0];

			Canvas.TextSize (BottomDataString[0], StringWidth, StringHeight);

			Canvas.SetPos
			(
				X + Slot_OffsetX_Right[0] - (StringWidth * 0.5f),
				Y + Cached.OffsetY_Bottom
			);

			Canvas.DrawText (BottomDataString[0], False);
			// </PacketLoss>


			// <Deaths>
			Canvas.DrawColor = BottomDataColor[1];

			Canvas.TextSize (BottomDataString[1], StringWidth, StringHeight);

			Canvas.SetPos
			(
				X + Slot_OffsetX_Right[1] - (StringWidth * 0.5f),
				Y + Cached.OffsetY_Bottom
			);

			Canvas.DrawText (BottomDataString[1], False);
			// </Deaths>


			// <PPH>
			Canvas.DrawColor = BottomDataColor[2];

			Canvas.TextSize (BottomDataString[2], StringWidth, StringHeight);

			Canvas.SetPos
			(
				X + Slot_OffsetX_Right[2] - (StringWidth * 0.5f),
				Y + Cached.OffsetY_Bottom
			);

			Canvas.DrawText (BottomDataString[2], False);
			// </PPH>
		}
		else
		{
			Canvas.Font = Cached.Font[SizeFont_Large];


			// <Kills>
			Canvas.DrawColor = TopDataColor[1];

			Canvas.TextSize (TopDataString[1], StringWidth, StringHeight);

			Canvas.SetPos
			(
				X + Slot_OffsetX_Right[1] - (StringWidth * 0.5f),
				Y + Cached.OffsetY_Top
			);

			Canvas.DrawText (TopDataString[1], False);
			// </Kills>


			// <Score>
			Canvas.DrawColor = TopDataColor[2];

			Canvas.TextSize (TopDataString[2], StringWidth, StringHeight);

			Canvas.SetPos
			(
				X + Slot_OffsetX_Right[2] - (StringWidth * 0.5f),
				Y + Cached.OffsetY_Top
			);

			Canvas.DrawText (TopDataString[2], False);
			// </Score>


			Canvas.Font = Cached.Font[SizeFont_Small];


			// <Deaths>
			Canvas.DrawColor = BottomDataColor[1];

			Canvas.TextSize (BottomDataString[1], StringWidth, StringHeight);

			Canvas.SetPos
			(
				X + Slot_OffsetX_Right[1] - (StringWidth * 0.5f),
				Y + Cached.OffsetY_Bottom
			);

			Canvas.DrawText (BottomDataString[1], False);
			// </Deaths>


			// <PPH>
			Canvas.DrawColor = BottomDataColor[2];

			Canvas.TextSize (BottomDataString[2], StringWidth, StringHeight);

			Canvas.SetPos
			(
				X + Slot_OffsetX_Right[2] - (StringWidth * 0.5f),
				Y + Cached.OffsetY_Bottom
			);

			Canvas.DrawText (BottomDataString[2], False);
			// </PPH>
		}
	}
	else
	{
		if (Level.NetMode == NM_Client)
		{
			Canvas.Font = Cached.Font[SizeFont_Large];


			// <DrawPing>
			Canvas.DrawColor = TopDataColor[0];

			Canvas.TextSize (TopDataString[0], StringWidth, StringHeight);

			Canvas.SetPos
			(
				X + Slot_OffsetX_Right[0] - (StringWidth * 0.5f),
				Y + Cached.OffsetY_Top
			);

			Canvas.DrawText (TopDataString[0], False);
			// </DrawPing>


			Canvas.Font = Cached.Font[SizeFont_Small];


			// <PacketLoss>
			Canvas.DrawColor = BottomDataColor[0];

			Canvas.TextSize (BottomDataString[0], StringWidth, StringHeight);

			Canvas.SetPos
			(
				X + Slot_OffsetX_Right[0] - (StringWidth * 0.5f),
				Y + Cached.OffsetY_Bottom
			);

			Canvas.DrawText (BottomDataString[0], False);
			// </PacketLoss>
		}
	}
}

// #

protected function SetPlayerStatusColor (Canvas Canvas, PlayerReplicationInfo OtherPRI)
{
	if (GRI.bMatchHasBegun)
	{
		if (RePRI(OtherPRI).IsCurrentPlayerStatus ('Out'))
		{
			Canvas.DrawColor = OutColor;
		}
		else if (RePRI(OtherPRI).IsCurrentPlayerStatus ('Dead'))
		{
			Canvas.DrawColor = DeadColor;
		}
		else
		{
			Canvas.DrawColor = LocationNameColor;
		}
    }
	else if (OtherPRI.bReadyToPlay)
	{
		Canvas.DrawColor = ReadyColor;
    }
    else
    {
		Canvas.DrawColor = WhiteColor;
    }
}

// #

protected function DrawEmptySlot
(
	Canvas Canvas,
	float X,
	float Y
)
{
	local float StringWidth, StringHeight;

	Canvas.DrawColor = RowColor;

	Canvas.SetPos (X, Y);

	Canvas.DrawTile
	(
		RowTexture,
		Cached.Width_Slot,
		Cached.Height_Slot,
		0.0f,
		0.0f,
		RowTexture.USize - 1.0f,
		RowTexture.VSize - 1.0f
	);

	Canvas.Font = Cached.Font[1];

	if (GRI.bMatchhasBegun)
	{
		if (Level.NetMode != NM_Standalone)
		{
			Canvas.DrawColor = BalanceTeamsColor;

			Canvas.StrLen (BalanceTeamsString, StringWidth, StringHeight);

			Canvas.SetPos
			(
				X + (Cached.Width_Slot * 0.5f) - (StringWidth * 0.5f),
				Y + (Cached.Height_Slot * 0.5f) - (StringHeight * 0.5f)
			);

			Canvas.DrawText (BalanceTeamsString, False);
		}
		else if (!PlayerOwnerPRI.bOnlySpectator)
		{
			Canvas.DrawColor = EmptySlotColor;

			Canvas.StrLen (SlotAvailableBotString, StringWidth, StringHeight);

			Canvas.SetPos
			(
				X + (Cached.Width_Slot * 0.5f) - (StringWidth * 0.5f),
				Y + (Cached.Height_Slot * 0.5f) - (StringHeight * 0.5f)
			);

			Canvas.DrawText (SlotAvailableBotString, False);
		}
		else
		{
			Canvas.DrawColor = EmptySlotColor;

			Canvas.StrLen (SlotAvailableBotString, StringWidth, StringHeight);

			Canvas.SetPos
			(
				X + (Cached.Width_Slot * 0.5f) - (StringWidth * 0.5f),
				Y + (Cached.Height_Slot * 0.5f) - (StringHeight * 0.5f)
			);

			Canvas.DrawText (SlotAvailableBotString, False);
		}
	}
	else if (Level.NetMode != NM_Standalone)
	{
		Canvas.DrawColor = EmptySlotColor;

		Canvas.StrLen (SlotAvailableString, StringWidth, StringHeight);

		Canvas.SetPos
		(
			X + (Cached.Width_Slot * 0.5f) - (StringWidth * 0.5f),
			Y + (Cached.Height_Slot * 0.5f) - (StringHeight * 0.5f)
		);

		Canvas.DrawText (SlotAvailableString, False);
	}
	else
	{
		Canvas.DrawColor = EmptySlotColor;

		Canvas.StrLen (SlotAvailableBotString, StringWidth, StringHeight);

		Canvas.SetPos
		(
			X + (Cached.Width_Slot * 0.5f) - (StringWidth * 0.5f),
			Y + (Cached.Height_Slot * 0.5f) - (StringHeight * 0.5f)
		);

		Canvas.DrawText (SlotAvailableBotString, False);
	}
}

// Preconditions:
// - (GRI.Teams[*] != None)

protected function float DrawTeamSummaries
(
	Canvas Canvas,
	float Y
)
{
	local string CachedString;
	local float StringWidth, StringHeight, Height_Bar, Y_Text;

	// Initialize

	Height_Bar = Cached.Height_Font[SizeFont_Small] * 1.5f;
	Y_Text = Y + (Height_Bar * 0.5f) - (Cached.Height_Font[SizeFont_Small] * 0.5f);

	// Your Team

	Canvas.DrawColor = Class'PlayerSettings'.Static.GetRelativeTeamColor (OwnTeamIndex, TeammateTeamIndex);
	Canvas.DrawColor.A = TeamSummaryBarOpacity;

	Canvas.SetPos (Cached.X_Slot[0], Y);

	Canvas.DrawTilePartialStretched
	(
		SummaryTexture,
		Cached.Width_Slot,
		Height_Bar
	);

	// Team Player Count

	Canvas.DrawColor = WhiteColor;

	Canvas.SetPos
	(
		Cached.X_Slot[0] + Cached.OffsetX_PlayerName,
		Y_Text
	);

	Canvas.DrawText (Teammates.Length @ "-" @ GRI.Teams[TeammateTeamIndex].GetHumanReadableName (), False);

	if (GRI.bMatchhasBegun)
	{
		// Team Kills Total

		CachedString = string(int(Metrics_Total_Score[TeammateTeamIndex]));

		Canvas.TextSize (CachedString, StringWidth, StringHeight);

		Canvas.SetPos
		(
			Cached.X_Slot[0] + Slot_OffsetX_Right[2] - (StringWidth * 0.5f),
			Y_Text
		);

		Canvas.DrawText (CachedString, False);

		// Team Score Average

		CachedString = string(Metrics_Total_Kills[TeammateTeamIndex]);

		Canvas.TextSize (CachedString, StringWidth, StringHeight);

		Canvas.SetPos
		(
			Cached.X_Slot[0] + Slot_OffsetX_Right[1] - (StringWidth * 0.5f),
			Y_Text
		);

		Canvas.DrawText (CachedString, False);
	}

	// Team Ping Average

	if (Level.NetMode == NM_Client)
	{
		Canvas.DrawColor = ConnectionColor;

		CachedString = string(Metrics_Average_Ping[TeammateTeamIndex]);

		Canvas.TextSize (CachedString, StringWidth, StringHeight);

		Canvas.SetPos
		(
			Cached.X_Slot[0] + Slot_OffsetX_Right[0] - (StringWidth * 0.5f),
			Y_Text
		);

		Canvas.DrawText (CachedString, False);
	}

	// Enemy Team

	Canvas.DrawColor = Class'PlayerSettings'.Static.GetRelativeTeamColor (OwnTeamIndex, EnemyTeamIndex);
	Canvas.DrawColor.A = TeamSummaryBarOpacity;

	Canvas.SetPos (Cached.X_Slot[1], Y);

	Canvas.DrawTilePartialStretched
	(
		SummaryTexture,
		Cached.Width_Slot,
		Height_Bar
	);

	// Team Player Count

	Canvas.DrawColor = WhiteColor;

	Canvas.SetPos
	(
		Cached.X_Slot[1] + Cached.OffsetX_PlayerName,
		Y_Text
	);

	Canvas.DrawText (Enemies.Length @ "-" @ GRI.Teams[EnemyTeamIndex].GetHumanReadableName (), False);

	if (GRI.bMatchhasBegun)
	{
		// Team Kills Total

		CachedString = string(int(Metrics_Total_Score[EnemyTeamIndex]));

		Canvas.TextSize (CachedString, StringWidth, StringHeight);

		Canvas.SetPos
		(
			Cached.X_Slot[1] + Slot_OffsetX_Right[2] - (StringWidth * 0.5f),
			Y_Text
		);

		Canvas.DrawText (CachedString, False);

		// Team Score Average

		CachedString = string(Metrics_Total_Kills[EnemyTeamIndex]);

		Canvas.TextSize (CachedString, StringWidth, StringHeight);

		Canvas.SetPos
		(
			Cached.X_Slot[1] + Slot_OffsetX_Right[1] - (StringWidth * 0.5f),
			Y_Text
		);

		Canvas.DrawText (CachedString, False);
	}

	// Team Ping Average

	if (Level.NetMode == NM_Client)
	{
		Canvas.DrawColor = ConnectionColor;

		CachedString = string(Metrics_Average_Ping[EnemyTeamIndex]);

		Canvas.TextSize (CachedString, StringWidth, StringHeight);

		Canvas.SetPos
		(
			Cached.X_Slot[1] + Slot_OffsetX_Right[0] - (StringWidth * 0.5f),
			Y_Text
		);

		Canvas.DrawText (CachedString, False);
	}

	// Advance Y

	return (Y + Height_Bar);
}

// Draws Player/Team versus Player/Team.

protected function float DrawVersusInfo (Canvas Canvas, float Y)
{
	local string Versus_L_Text_Left, Versus_L_Text_Right;

	local float Versus_L_X,
		Versus_L_Y,
		Versus_L_Width_Left,
		Versus_L_Width_Right,
		Versus_L_Height_Text,
		Versus_L_Width_Background,
		Versus_L_Height_Background,
		Versus_L_Width_Symbol,
		Versus_L_Height_Symbol,
		Versus_L_Ratio;

	//

	Canvas.Font = HUDClass.Static.GetMediumFontfor (Canvas);
	Canvas.Font = GetSmallerFontfor (Canvas, 0);

	Canvas.TextSize ("*", Versus_L_Width_Symbol, Versus_L_Height_Symbol);

	Versus_L_Width_Background = Canvas.ClipX * Bar_ScaledWidth;
	Versus_L_Height_Background = Versus_L_Height_Symbol * 1.25f;

	Canvas.SetPos
	(
		Canvas.ClipX * Bar_ScaledPositionX,
		Y
	);


	if (Teammates.Length == 1)
	{
		Versus_L_Text_Left = Teammates[0].PlayerName;
	}
	else
	{
		Versus_L_Text_Left = GRI.Teams[TeammateTeamIndex].GetHumanReadableName ();
	}

	if (Enemies.Length == 1)
	{
		Versus_L_Text_Right = Enemies[0].PlayerName;
	}
	else
	{
		Versus_L_Text_Right = GRI.Teams[EnemyTeamIndex].GetHumanReadableName ();
	}

	Canvas.TextSize (Versus_L_Text_Left, Versus_L_Width_Left, Versus_L_Height_Text);
	Canvas.TextSize (Versus_L_Text_Right, Versus_L_Width_Right, Versus_L_Height_Text);

	// Background Icon

	Versus_L_Ratio = (Versus_L_Height_Text * 0.5f) / (VersusSymbolOffset.Y2 - VersusSymbolOffset.Y1);

	Versus_L_Width_Symbol = (VersusSymbolOffset.X2 - VersusSymbolOffset.X1) * Versus_L_Ratio;
	Versus_L_Height_Symbol = (VersusSymbolOffset.Y2 - VersusSymbolOffset.Y1) * Versus_L_Ratio;

	Versus_L_X = Canvas.ClipX * 0.5f;
	Versus_L_Y = Y + (Versus_L_Height_Background * 0.5f) - (Versus_L_Height_Symbol * 0.5f);

	Canvas.DrawColor = VersusSymbolColor;

	Canvas.SetPos
	(
		Versus_L_X - (Versus_L_Width_Symbol * 0.5f),
		Versus_L_Y
	);

	Canvas.DrawTile
	(
		VersusSymbolTexture,
		Versus_L_Width_Symbol,
		(VersusSymbolOffset.Y2 - VersusSymbolOffset.Y1) * Versus_L_Ratio,
		VersusSymbolOffset.X1,
		VersusSymbolOffset.Y1,
		VersusSymbolOffset.X2 - VersusSymbolOffset.X1,
		VersusSymbolOffset.Y2 - VersusSymbolOffset.Y1
	);

	//

	Versus_L_Y = Y + (Versus_L_Height_Background * 0.5f) - (Versus_L_Height_Text * 0.5f);

	// Left

	Canvas.DrawColor = TeammateColor;
	Canvas.DrawColor.A = NameOpacity;

	Canvas.SetPos
	(
		Versus_L_X - Versus_L_Width_Left - Versus_L_Height_Text,
		Versus_L_Y
	);

	Canvas.DrawText (Versus_L_Text_Left, False);

	// Right

	Canvas.DrawColor = EnemyColor;
	Canvas.DrawColor.A = NameOpacity;

	Canvas.SetPos
	(
		Versus_L_X + Versus_L_Height_Text,
		Versus_L_Y
	);

	Canvas.DrawText (Versus_L_Text_Right, False);

	// Advanced Y

	return Y + Versus_L_Height_Background;
}

// Draws YouveWonTheMatch, YouveLostTheMatch, Ended, ...

protected function float DrawMatchStatus (Canvas Canvas, float Y)
{
	local ReGRI ReGRI;
	local string CachedString;
	local float Width_Symbol;
	local float Height_Symbol;

	ReGRI = ReGRI(GRI);

	Canvas.Font = PlayerOwnerHUD.GetFontSizeIndex (Canvas, 2);

	if (bMatchOver)
	{
		if (UnrealPlayer(Owner).bDisplayWinner)
		{
			Canvas.DrawColor = Class'PlayerSettings'.Static.GetRelativeTeamColor (OwnTeamIndex, TeammateTeamIndex);
			Canvas.DrawColor.A = MatchStatusOpacity;

			Canvas.TextSize (Class'XInterface.HUDBase'.Default.YouveWonTheMatch, Width_Symbol, Height_Symbol);

			Canvas.SetPos
			(
				(Canvas.ClipX * 0.5f) - (Width_Symbol * 0.5f),
				Y
			);

			Canvas.DrawText (Class'XInterface.HUDBase'.Default.YouveWonTheMatch, False);

			return (Y + Height_Symbol);
		}
		else if (UnrealPlayer(Owner).bDisplayLoser)
		{
			Canvas.DrawColor = Class'PlayerSettings'.Static.GetRelativeTeamColor (OwnTeamIndex, TeammateTeamIndex);
			Canvas.DrawColor.A = MatchStatusOpacity;

			Canvas.TextSize (Class'XInterface.HUDBase'.Default.YouveLostTheMatch, Width_Symbol, Height_Symbol);

			Canvas.SetPos
			(
				(Canvas.ClipX * 0.5f) - (Width_Symbol * 0.5f),
				Y
			);

			Canvas.DrawText (Class'XInterface.HUDBase'.Default.YouveLostTheMatch, False);

			return (Y + Height_Symbol);
		}
		else if (PlayerReplicationInfo(GRI.Winner) != None)
		{
			Canvas.DrawColor = WhiteColor;
			Canvas.DrawColor.A = MatchStatusOpacity;

			CachedString = Class'XInterface.HUDBase'.Default.WonMatchPrefix $ PlayerReplicationInfo(GRI.Winner).PlayerName $ Class'XInterface.HUDBase'.Default.WonMatchPostfix;

			Canvas.TextSize (CachedString, Width_Symbol, Height_Symbol);

			Canvas.SetPos
			(
				(Canvas.ClipX * 0.5f) - (Width_Symbol * 0.5f),
				Y
			);

			Canvas.DrawText (CachedString, False);

			return (Y + Height_Symbol);
		}
		else if (TeamInfo(GRI.Winner) != None)
		{
			Canvas.DrawColor = Class'PlayerSettings'.Default.StandardTeamColor[TeamInfo(GRI.Winner).TeamIndex];
			Canvas.DrawColor.A = MatchStatusOpacity;

			CachedString = Class'XInterface.HUDBase'.Default.WonMatchPrefix $ TeamInfo(GRI.Winner).GetHumanReadableName () $ Class'XInterface.HUDBase'.Default.WonMatchPostfix;

			Canvas.TextSize (CachedString, Width_Symbol, Height_Symbol);

			Canvas.SetPos
			(
				(Canvas.ClipX * 0.5f) - (Width_Symbol * 0.5f),
				Y
			);

			Canvas.DrawText (CachedString, False);

			return (Y + Height_Symbol);
		}
		else
		{
			Canvas.DrawColor = WhiteColor;
			Canvas.DrawColor.A = MatchStatusOpacity;

			Canvas.TextSize (Ended, Width_Symbol, Height_Symbol);

			Canvas.SetPos
			(
				(Canvas.ClipX * 0.5f) - (Width_Symbol * 0.5f),
				Y
			);

			Canvas.DrawText (Ended, False);

			return (Y + Height_Symbol);
		}
	}
	else
	{
		Canvas.TextSize ("*", Width_Symbol, Height_Symbol);

		return (Y + Height_Symbol);
	}
}

// #

DefaultProperties
{
	ScaledWidth_Scoreboard = 0.96
	ScaledWidth_Slot = 0.47
	ScaledPositionX_Scoreboard = 0.02
	ScaledPositionY_Scoreboard = 0.125
	ScaledPositionX_Slot(0) = 0.02
	ScaledPositionX_Slot(1) = 0.51

	// Strings

	PlayerString = "Player"
	SlotAvailableBotString = "Slot Available (Bot)"
	SlotAvailableString = "Slot Available"
	BalanceTeamsString = "Balance Teams"

	// Colors

	TeamNeutralColor = (R=223,G=223,B=223,A=255)
	BalanceTeamsColor = (R=255,G=255,B=0,A=150)
	EmptySlotColor = (R=159,G=159,B=159,A=255)
	CountryFlagColor = (R=255,G=255,B=255,A=191)
	LegendColor = (R=255,G=255,B=255,A=255)
	RowColor = (R=255,G=255,B=255,A=191)
	AdministratorColor = (R=120,G=36,B=172,A=180)
	PortraitColor = (R=255,G=255,B=255,A=255)
	CountryCodeColor = (R=178,G=178,B=0,A=255)

	// Opacity

	TeamScoreBarOpacity = 223
	TeamBarOpacity = 223
	TeamSummaryBarOpacity = 223
	NameOpacity = 150
	OwnNameOpacity = 191
	MatchStatusOpacity = 223

	// Legend

	TopLegendColor(0) = (R=20,G=127,B=150,A=255)
	TopLegendString(0) = "Ping"
	BottomLegendColor(0) = (R=20,G=127,B=150,A=255)
	BottomLegendString(0) = "P/L"
	EmptyTopString(0) = "---"

	TopLegendColor(1) = (R=178,G=178,B=178,A=255)
	TopLegendString(1) = "Kills"
	BottomLegendColor(1) = (R=198,G=20,B=20,A=255)
	BottomLegendString(1) = "Deaths"
	EmptyTopString(1) = "---"

	TopLegendColor(2) = (R=178,G=178,B=178,A=255)
	TopLegendString(2) = "Score"
	BottomLegendColor(2) = (R=178,G=178,B=178,A=255)
	BottomLegendString(2) = "PPH"
	EmptyTopString(2) = "---"

	TopLegendColor(3) = (R=178,G=178,B=178,A=255)
	TopLegendString(3) = ""
	BottomLegendColor(3) = (R=178,G=178,B=178,A=255)
	BottomLegendString(3) = ""
	EmptyTopString(3) = ""

	//

	LegendTexture = Texture'Textures.PlayerRow'
	TeamScoreTexture = Texture'Textures.ScoreboardTeamLegend'
	RowTexture = Texture'Textures.PlayerRow'
	OwnRowTexture = Texture'Textures.ScoreboardPlayerSlot'
	SummaryTexture = Texture'Textures.ScoreboardTeamLegend'

	// Scoreboard

	HUDClass = Class'ReTDM_HUD'
}
