// *F* Developed 2011-2014 rejecht <rejecht at outlook dot com>

class ReUtility extends Object
	abstract
	HideDropDown
	;

const IndexTeam_Max = 2;
const IndexTeam_A = 0;
const IndexTeam_B = 1;
const IndexTeam_Unspecified = 255;

// - Determine whether the player (PRI) is local.
// - GetLocalPlayerController is simulated client-side.

static final function bool IsLocalPlayer (PlayerReplicationInfo PRI)
{
	local PlayerController PC;

	if (PRI != None)
	{
		PC = PRI.Level.GetLocalPlayerController ();

		return
		(
			(PC != None)
			&&
			(PC.PlayerReplicationInfo == PRI)
		);
	}

	return false;
}

// ## AdrenalineCombo

// Preconditions:
// - (Class_Combo != None)
// Protocol:
// - Return "????" if the combo was not found.
// - Return uppercase B (Back), F (Forward), L (Left), R (Right).

static final function string GetKeyString (byte InputKey)
{
	if (InputKey == 1)
	{
		return "F";
	}
	else if (InputKey == 2)
	{
		return "B";
	}
	else if (InputKey == 4)
	{
		return "R";
	}
	else if (InputKey == 8)
	{
		return "L";
	}
	else
	{
		return "?";
	}
}

static final function string GetKeyComboString (class<Combo> ComboClass)
{
	return
	(
		GetKeyString (ComboClass.Default.Keys[3])
		$ GetKeyString (ComboClass.Default.Keys[2])
		$ GetKeyString (ComboClass.Default.Keys[1])
		$ GetKeyString (ComboClass.Default.Keys[0])
	);
}

// Protocol:
// - Return false if could not be determined.
// - Return false if not a teamkill.
// - Return true if and only if a teamkill.

static final function bool IsTeamKill
(
	PlayerReplicationInfo Killer,
	PlayerReplicationInfo Other
)
{
	return
	(
		(Killer != None)
		&&
		(Killer.Level.GRI != None)
		&&
		(Killer.Level.GRI.bTeamGame)
		&&
		(Other != None)
		&&
		(Killer.Team != None)
		&&
		(Other.Team != None)
		&&
		(Killer.Team == Other.Team)
	);
}

// - To provide a color-coded player name.
// -- A local player has a distinct color (yellow).
// -- An opponent has a distinct color (white, red, blue).
// - If PRI == None, it returns the Default "Someone" name.

static final function string GetColorCodedPlayerNameString
(
	PlayerReplicationInfo PRI,
	optional bool bPreviousName
)
{
	local PlayerController PC;
	local string ColorCodedPlayerNameString;

	if (PRI == None)
	{
		return
		(
			GetColorCodeString (Class'PlayerSettings'.Default.EnemyPlayerColor)
			$ class'XGame.xDeathMessage'.Default.SomeoneString
		);
	}
	else if (PRI.PlayerName == "")
	{
		return
		(
			GetColorCodeString (Class'PlayerSettings'.Default.EnemyPlayerColor)
			$ class'XGame.xDeathMessage'.Default.SomeoneString
		);
	}

	// This is assumed successful and not checked.

	PC = PRI.Level.GetLocalPlayerController ();

	if (PRI.Team != None)
	{
		if (PRI.Team.TeamIndex == IndexTeam_A)
		{
			if (!IsLocalPlayer (PRI))
			{
				ColorCodedPlayerNameString = Class'PlayerSettings'.static.GetRelativePlayerColorCodedString (PC.GetTeamNum (), IndexTeam_A);
			}
			else
			{
				ColorCodedPlayerNameString = GetColorCodeString (Class'PlayerSettings'.Default.CustomPlayerColor);
			}
		}
		else if (PRI.Team.TeamIndex == IndexTeam_B)
		{
			if (!IsLocalPlayer (PRI))
			{
				ColorCodedPlayerNameString = Class'PlayerSettings'.static.GetRelativePlayerColorCodedString (PC.GetTeamNum (), IndexTeam_B);
			}
			else
			{
				ColorCodedPlayerNameString = GetColorCodeString (Class'PlayerSettings'.Default.CustomPlayerColor);
			}
		}
		else if (PRI.Team.TeamIndex == IndexTeam_Unspecified)
		{
			ColorCodedPlayerNameString = Class'PlayerSettings'.static.GetRelativePlayerColorCodedString (PC.GetTeamNum (), IndexTeam_Unspecified);
		}
		else
		{
			ColorCodedPlayerNameString = Class'PlayerSettings'.static.GetRelativePlayerColorCodedString (PC.GetTeamNum (), IndexTeam_Unspecified);
		}
	}
	else if (!IsLocalPlayer (PRI)) // Free-for-all local player.
	{
		ColorCodedPlayerNameString = GetColorCodeString (Class'PlayerSettings'.Default.EnemyPlayerColor);
	}
	else // Free-for-all other players.
	{
		ColorCodedPlayerNameString = GetColorCodeString (Class'PlayerSettings'.Default.CustomPlayerColor);
	}

	if (!bPreviousName)
	{
		ColorCodedPlayerNameString $= PRI.PlayerName;
	}
	else
	{
		ColorCodedPlayerNameString $= PRI.OldName;
	}

	return ColorCodedPlayerNameString;
}

// - The primary reason for referencing xDeathMessage is to reduce the
// number of strings to localize.
// - bPreviousName was added later to account for some contexts where the old name
// is required (name change).

static final function string GetPlayerNameString
(
	PlayerReplicationInfo PRI,
	optional bool bPreviousName
)
{
	if (PRI == None)
	{
		return class'XGame.xDeathMessage'.Default.SomeoneString;
	}

	if (!bPreviousName)
	{
		if (PRI.PlayerName != "")
		{
			return PRI.PlayerName;
		}
		else
		{
			return class'XGame.xDeathMessage'.Default.SomeoneString;
		}
	}
	else
	{
		if (PRI.OldName != "")
		{
			return PRI.OldName;
		}
		else
		{
			return class'XGame.xDeathMessage'.Default.SomeoneString;
		}
	}
}

static final function Color GetPlayerColor (PlayerReplicationInfo OtherPRI)
{
	local PlayerController PC;

	if (OtherPRI != None)
	{
		PC = OtherPRI.Level.GetLocalPlayerController ();

		if (PC != None)
		{
			if (!PC.PlayerReplicationInfo.bOnlySpectator)
			{
				if (PC.PlayerReplicationInfo != OtherPRI)
				{
					if (OtherPRI.Team != None)
					{
						return Class'PlayerSettings'.Static.GetRelativeTeamColor (PC.GetTeamNum (), OtherPRI.Team.TeamIndex);
		            }
		            else
		            {
						return Class'PlayerSettings'.Default.EnemyPlayerColor;
		            }
	            }
	            else
	            {
					return Class'PlayerSettings'.Default.CustomPlayerColor;
	            }
	        }
	        else if (OtherPRI.Team != None)
	        {
				return Class'PlayerSettings'.Default.StandardTeamColor[OtherPRI.Team.TeamIndex];
	        }
			else if (PC.PlayerReplicationInfo != OtherPRI)
			{
				return Class'PlayerSettings'.Default.EnemyPlayerColor;
			}
			else
			{
				return Class'PlayerSettings'.Default.CustomPlayerColor;
			}
		}
		else
		{
			return Class'PlayerSettings'.Default.UnknownPlayerColor;
        }
    }
    else
    {
		return Class'PlayerSettings'.Default.UnknownPlayerColor;
    }
}

static final function Color GetTeamColor (PlayerReplicationInfo RelatedPRI_1)
{
	local PlayerController PC;

	if (RelatedPRI_1 != None)
	{
		PC = RelatedPRI_1.Level.GetLocalPlayerController ();

		if
		(
			(PC != None)
			&&
			(PC.DemoViewer == None)
		)
		{
			if (!PC.PlayerReplicationInfo.bOnlySpectator) // Relative Colors
			{
				if (RelatedPRI_1.Team != None)
				{
					return Class'PlayerSettings'.Static.GetRelativeTeamColor (PC.GetTeamNum (), RelatedPRI_1.Team.TeamIndex);
	            }
	            else
	            {
					return Class'PlayerSettings'.Default.EnemyPlayerColor;
	            }
	        }
	        else if (RelatedPRI_1.Team != None) // Absolute Colors
	        {
				return Class'PlayerSettings'.Default.StandardTeamColor[RelatedPRI_1.Team.TeamIndex];
	        }
			else
			{
				return Class'PlayerSettings'.Default.SpectatorSayColor;
			}
		}
        else if (RelatedPRI_1.Team != None) // Absolute Colors
        {
			return Class'PlayerSettings'.Default.StandardTeamColor[RelatedPRI_1.Team.TeamIndex];
        }
		else
		{
			return Class'PlayerSettings'.Default.UnknownPlayerColor;
        }
    }
    else
    {
		return Class'PlayerSettings'.Default.UnknownPlayerColor;
    }
}

static final function bool IsHeadShot (Class<DamageType> Class_DamageType)
{
	return
	(
		(Class_DamageType == class'XWeapons.DamTypeSniperHeadShot')
		||
		(Class_DamageType == class'UTClassic.DamTypeClassicHeadShot')
	);
}

// - To return a color coded text string with the text representation
// of the damage type.
// (bGeneralizedWeaponDamage) show name of weapon rather than ammunition.
// Preconditions:
// - (Class_DamageType != None)

static final function string GetDamageTypeColorCodedString
(
	Class<DamageType> DamageTypeClass
)
{
	local int Index;
	local int Count;

	if (ClassIsChildof (DamageTypeClass, Class'UnrealGame.WeaponDamageType'))
	{
		if (Class'PlayerSettings'.Default.DeathMessage_bUseWeaponColor)
		{
			if (IsHeadShot (DamageTypeClass))
			{
				if (Class'PlayerSettings'.Default.DeathMessage_bUseWeaponName)
				{
					return
					(
						GetColorCodeString (Class'PlayerSettings'.Default.HeadshotDamageTypeColor)
						$ "[HS]"
						@ GetColorCodeString (Class<WeaponDamageType>(DamageTypeClass).Default.WeaponClass.Default.HUDColor)
						$ "["
						$ Class<WeaponDamageType>(DamageTypeClass).Default.WeaponClass.Default.ItemName
						$ "]"
					);
				}
				else
				{
					return
					(
						GetColorCodeString (Class'PlayerSettings'.Default.HeadshotDamageTypeColor)
						$ "[HS]"
						@ GetColorCodeString (Class<WeaponDamageType>(DamageTypeClass).Default.WeaponClass.Default.HUDColor)
						$ "["
						$ Class<WeaponDamageType>(DamageTypeClass).Default.WeaponClass.Default.FireModeClass[0].Default.AmmoClass.Default.ItemName
						$ "]"
					);
				}
			}
			else // Is not a head shot.
			{
				if (Class'PlayerSettings'.Default.DeathMessage_bUseWeaponName)
				{
					return
					(
						GetColorCodeString (Class<WeaponDamageType>(DamageTypeClass).Default.WeaponClass.Default.HUDColor)
						$ "["
						$ Class<WeaponDamageType>(DamageTypeClass).Default.WeaponClass.Default.ItemName
						$ "]"
					);
				}
				else // Is not generalized.
				{
					return
					(
						GetColorCodeString (Class<WeaponDamageType>(DamageTypeClass).Default.WeaponClass.Default.HUDColor)
						$ "["
						$ Class<WeaponDamageType>(DamageTypeClass).Default.WeaponClass.Default.FireModeClass[0].Default.AmmoClass.Default.ItemName
						$ "]"
					);
				}
			}
		}
		else
		{
			if (IsHeadShot (DamageTypeClass))
			{
				if (Class'PlayerSettings'.Default.DeathMessage_bUseWeaponName)
				{
					return
					(
						GetColorCodeString (Class'PlayerSettings'.Default.HeadshotDamageTypeColor)
						$ "[HS]"
						@ GetColorCodeString (Class'PlayerSettings'.Default.WeaponDamageTypeColor)
						$ "["
						$ Class<WeaponDamageType>(DamageTypeClass).Default.WeaponClass.Default.ItemName
						$ "]"
					);
				}
				else // Is not generalized.
				{
					return
					(
						GetColorCodeString (Class'PlayerSettings'.Default.HeadshotDamageTypeColor)
						$ "[HS]"
						@ GetColorCodeString (Class'PlayerSettings'.Default.WeaponDamageTypeColor)
						$ "["
						$ Class<WeaponDamageType>(DamageTypeClass).Default.WeaponClass.Default.FireModeClass[0].Default.AmmoClass.Default.ItemName
						$ "]"
					);
				}
			}
			else
			{
				if (Class'PlayerSettings'.Default.DeathMessage_bUseWeaponName)
				{
					return
					(
						GetColorCodeString (Class'PlayerSettings'.Default.WeaponDamageTypeColor)
						$ "["
						$ Class<WeaponDamageType>(DamageTypeClass).Default.WeaponClass.Default.ItemName
						$ "]"
					);
				}
				else // Is not generalized.
				{
					return
					(
						GetColorCodeString (Class'PlayerSettings'.Default.WeaponDamageTypeColor)
						$ "["
						$ Class<WeaponDamageType>(DamageTypeClass).Default.WeaponClass.Default.FireModeClass[0].Default.AmmoClass.Default.ItemName
						$ "]"
					);
				}
			}
		}
	}
	else
	{
		Count = Class'DamageTypeMap'.Default.DamageTypeTranslation.Length;

		for (Index = 0; Index < Count; ++Index)
		{
			if (Class'DamageTypeMap'.Default.DamageTypeTranslation[Index].DamageTypeClass != DamageTypeClass)
			{
				continue;
			}
			else
			{
				return
				(
					GetColorCodeString (Class'PlayerSettings'.Default.GeneralDamageTypeColor)
					$ "["
					$ Class'DamageTypeMap'.Default.DamageTypeTranslation[Index].DamageTypeString
					$ "]"
				);
            }
		}
	}

	return GetUnhandledDamageTypeColorCodedString (DamageTypeClass);
}

// Just return the text.
// Preconditions:
// - (Class_DamageType != None)

private static final function string GetUnhandledDamageTypeColorCodedString
(
	Class<DamageType> DamageTypeClass
)
{
	local string DamageTypeString;
	local int Index;

	DamageTypeString = string(DamageTypeClass);
	Index = InStr (DamageTypeString, ".");

	if (Index > 1)
	{
		return
		(
			GetColorCodeString (Class'PlayerSettings'.Default.UnknownDamageTypeColor)
			$ "["
			$ Mid (DamageTypeString, Index + 1)
			$ "]"
		);
	}

	return
	(
		GetColorCodeString (Class'PlayerSettings'.Default.UnknownDamageTypeColor)
		$ "["
		$ DamageTypeClass
		$ "]"
	);
}

static final function string GetUnhandledDamageTypeString (Class<DamageType> DamageTypeClass)
{
	local string DamageTypeString;
	local int Index;

	DamageTypeString = string(DamageTypeClass);
	Index = InStr (DamageTypeString, "."); // Skips package name

	if (Index > 1)
	{
		return Mid (DamageTypeString, Index + 1);
	}

	return string(DamageTypeClass);
}

// #

static final function string GetDamageTypeString (Class<DamageType> DamageTypeClass)
{
	local int Index;
	local int Count;

	if (ClassIsChildof (DamageTypeClass, Class'UnrealGame.WeaponDamageType'))
	{
		if (Class'PlayerSettings'.Default.DeathMessage_bUseWeaponName)
		{
			return Class<WeaponDamageType>(DamageTypeClass).Default.WeaponClass.Default.ItemName;
		}
		else // Is not generalized.
		{
			return Class<WeaponDamageType>(DamageTypeClass).Default.WeaponClass.Default.FireModeClass[0].Default.AmmoClass.Default.ItemName;
		}
	}
	else
	{
		Count = Class'DamageTypeMap'.Default.DamageTypeTranslation.Length;

		for (Index = 0; Index < Count; ++Index)
		{
			if (Class'DamageTypeMap'.Default.DamageTypeTranslation[Index].DamageTypeClass != DamageTypeClass)
			{
				continue;
			}
			else
			{
				return Class'DamageTypeMap'.Default.DamageTypeTranslation[Index].DamageTypeString;
            }
		}
	}

	return GetUnhandledDamageTypeString (DamageTypeClass);
}

// - Text string are zero-terminated.
// - Color code format is four binary octets.
// -- 0x1B is the prefix (escape code from ANSI).
// --- Red: [0x01, 0xFF]
// --- Green: [0x01, 0xFF]
// --- Blue: [0x01, 0xFF]

static final function string GetColorCodeString (Color Color)
{
	return
	(
		Chr (0x1B)
		$ Chr (Max (Color.R, 1))
		$ Chr (Max (Color.G, 1))
		$ Chr (Max (Color.B, 1))
	);
}

static final function bool SkipCharacterCode (int Code)
{
	return
	(
		(Code < 0x20) // Controls C0
		||
		(
			(Code > 0x7E) // Controls C0 [0x7F (127), 0xA0 (160)]
			&&
			(Code < 0xA1) // Controls A0 + Non-breaking space (0xA0), not < C1
		)
	);
}

static final function string GetPlainString (coerce string SourceString)
{
	local string DestinationString;
	local string CharacterString;
	local int Index;
	local int Count_Characters;


    Index = InStr (SourceString, Chr (0x1B)); // First Color Code Prefix


	while (Index > -1)
	{
		SourceString = Left (SourceString, Index) $ Mid (SourceString, Index + 4);
		Index = InStr (SourceString, Chr (0x1B)); // Next Color Code Prefix
	}


	Count_Characters = Len (SourceString);


	for (Index = 0; Index < Count_Characters; ++Index)
	{
		CharacterString = Mid (SourceString, Index, 1);

		if (!SkipCharacterCode (Asc (CharacterString)))
		{
			DestinationString $= CharacterString;
		}
	}


	return DestinationString;
}

// - Follows ISO 8601 format YYYY-MM-DD hh:mm:ss.
// - Reminder: Eval is at least five times slower than using if-else checks.

static final function string GetLocalTimeString (LevelInfo LI)
{
	local string CachedString;

	CachedString = LI.Year$"-";

	if (LI.Month < 10)
	{
		CachedString $= "0"$LI.Month$"-";
	}
	else
	{
		CachedString $= LI.Month$"-";
	}

	if (LI.Day > 9)
	{
		CachedString $= LI.Day$" ";
	}
	else
	{
		CachedString $= "0"$LI.Day$" ";
	}

	if (LI.Hour > 9)
	{
		CachedString $= LI.Hour$":";
	}
	else
	{
		CachedString $= "0"$LI.Hour$":";
	}

	if (LI.Minute > 9)
	{
		CachedString $= LI.Minute$":";
	}
	else
	{
		CachedString $= "0"$LI.Minute$":";
	}

	if (LI.Second > 9)
	{
		CachedString $= LI.Second;
	}
	else
	{
		CachedString $= "0"$LI.Second;
	}

	return CachedString;
}

static final function string GetText_HhMmSs (LevelInfo LI)
{
	local string CachedString;

	if (LI.Hour > 9)
	{
		CachedString = LI.Hour$":";
	}
	else
	{
		CachedString = "0"$LI.Hour$":";
	}

	if (LI.Minute > 9)
	{
		CachedString $= LI.Minute$":";
	}
	else
	{
		CachedString $= "0"$LI.Minute$":";
	}

	if (LI.Second > 9)
	{
		CachedString $= LI.Second;
	}
	else
	{
		CachedString $= "0"$LI.Second;
	}

	return CachedString;
}

// - Returns "Mm:Ss.us"
// - Useful for logging quick events during development.

static final function string GetShortTimeString (LevelInfo LI)
{
	local string CachedString;

	if (LI.Minute > 9)
	{
		CachedString = LI.Minute $ ":";
	}
	else
	{
		CachedString = "0" $ LI.Minute $ ":";
	}


	if (LI.Second > 9)
	{
		CachedString $= LI.Second $ ".";
	}
	else
	{
		CachedString $= "0" $ LI.Second $ ".";
	}


	if (LI.Millisecond > 99)
	{
		CachedString $= LI.Millisecond;
	}
	else if (LI.Millisecond > 9)
	{
		CachedString $= "0" $ LI.Millisecond;
	}
	else
	{
		CachedString $= "00" $ LI.Millisecond;
	}


	return CachedString;
}

// - Returns an ISO 8601 format YYYYMMDDThhmmss string.
// - Reminder: Eval is at least five times slower than using if-else checks.

static final function string GetCompressedDatestampString (LevelInfo LI)
{
	local string CachedString;

	CachedString = string(LI.Year);

	if (LI.Month < 10)
	{
		CachedString $= "0" $ LI.Month;
	}
	else
	{
		CachedString $= LI.Month;
	}

	if (LI.Day > 9)
	{
		CachedString $= LI.Day $ "T";
	}
	else
	{
		CachedString $= "0" $ LI.Day $ "T";
	}

	if (LI.Hour > 9)
	{
		CachedString $= LI.Hour;
	}
	else
	{
		CachedString $= "0" $ LI.Hour;
	}

	if (LI.Minute > 9)
	{
		CachedString $= LI.Minute;
	}
	else
	{
		CachedString $= "0" $ LI.Minute;
	}

	if (LI.Second > 9)
	{
		CachedString $= LI.Second;
	}
	else
	{
		CachedString $= "0" $ LI.Second;
	}

	return CachedString;
}

// Returns:
// - m:Ss
// - Mm:Ss
// - h:Mm:Ss
// - Hh:Mm:Ss

static final function string GetText_Countdown_HhMmSs (int Seconds)
{
	local string CachedString;
	local int Minutes;
	local int Hours;

	if (Seconds < 3600) // Returns (m:Ss || Mm:Ss)
	{
		Minutes = Seconds / 60;
		Seconds -= Minutes * 60;

		if (Seconds > 9)
		{
			return (Minutes $ ":" $ Seconds);
        }
        else
        {
			return (Minutes $ ":0" $ Seconds);
        }
	}
	else if (Seconds < 356400) // Returns (h:Mm:Ss || Hh:Mm:Ss)
	{
		Hours = Seconds / 3600;
		Seconds -= Hours * 3600;
		Minutes = Seconds / 60;
		Seconds -= Minutes * 60;

		if (Minutes > 9)
		{
			CachedString = Hours $ ":" $ Minutes;
        }
        else
        {
			CachedString = Hours $ ":0" $ Minutes;
        }

		if (Seconds > 9)
		{
			return CachedString $ ":" $ Seconds;
        }
        else
        {
			return CachedString $ ":0" $ Seconds;
        }
	}
}

// ## Development ##

static final function string GetNetworkModeString (LevelInfo.ENetMode NetMode)
{
	switch (NetMode)
	{
		case NM_Standalone:
			return "NM_Standalone (0)";

		case NM_DedicatedServer:
			return "NM_DedicatedServer (1)";

		case NM_ListenServer:
			return "NM_ListenServer (2)";

		case NM_Client:
			return "NM_Client (3)";
	}

	return "<NM_Unknown ("$NetMode$")>";
}

static final function string GetNetworkRoleString (Actor.ENetRole NetworkRole)
{
	switch (NetworkRole)
	{
		case ROLE_None:
			return "ROLE_None (0)";

		case ROLE_DumbProxy:
			return "ROLE_DumbProxy (1)";

		case ROLE_SimulatedProxy:
			return "ROLE_SimulatedProxy (2)";

		case ROLE_AutonomousProxy:
			return "ROLE_AutonomousProxy (3)";

		case ROLE_Authority:
			return "ROLE_Authority (4)";
	}

	return "<ROLE_Unknown ("$NetworkRole$")>";
}

static final function string GetPhysicsString (Actor.EPhysics Index)
{
	switch (Index)
	{
		case PHYS_None:
			return "PHYS_None";

		case PHYS_Walking:
			return "PHYS_Walking";

		case PHYS_Falling:
			return "PHYS_Falling";

		case PHYS_Swimming:
			return "PHYS_Swimming";

		case PHYS_Flying:
			return "PHYS_Flying";

		case PHYS_Rotating:
			return "PHYS_Rotating";

		case PHYS_Projectile:
			return "PHYS_Projectile";

		case PHYS_Interpolating:
			return "PHYS_Interpolating";

		case PHYS_MovingBrush:
			return "PHYS_MovingBrush";

		case PHYS_Spider:
			return "PHYS_Spider";

		case PHYS_Trailer:
			return "PHYS_Trailer";

		case PHYS_Ladder:
			return "PHYS_Ladder";

		case PHYS_RootMotion:
			return "PHYS_RootMotion";

		case PHYS_Karma:
			return "PHYS_Karma";

		case PHYS_KarmaRagDoll:
			return "PHYS_KarmaRagDoll";

		case PHYS_Hovering:
			return "PHYS_Hovering";

		case PHYS_CinMotion:
			return "PHYS_CinMotion";
	}

	return "<PHYS_Unknown ("$Index$")>";
}

// ####

DefaultProperties
{
}
